#!python
#cython: language_level=3
#define NPY_NO_DEPRECATED_API NPY_1_7_API_VERSION

"""
Created on Tue Apr  5 10:27:28 2022

This code uses Cython to speed up multiple functions commonly used in Smuthi.
The code also has functions for calculating the explicit direct particle-particle
coupling matrix within a single layer. A hash table method is used to speed 
up the calculation of the explicit direct coupling matrix in both the 2D and 3D 
case. T

NOTE: The code above this comment section is important. 
cython: language_level=3 -> This enforces python3 standards.
Most importantly 1/3 = 0.33333. language_level=2 -> 1/3 = 0

@author: parkerwray
"""

import scipy.special
import numpy as np

# python imports before cython imports, so cython overwrites.
cimport scipy.special.cython_special
cimport numpy as np
cimport cython
from cython.parallel import prange

## Note: All C libraries can be found for Cython in https://github.com/cython/cython/tree/master/Cython/Includes
cdef extern from "complex.h" nogil:
    double complex cexp(double complex)
    double complex cpow(double complex x, double complex y)
    double complex csqrt(double complex z)
    double cabs(double complex z)
    double cimag(double complex z)
    double creal(double complex z)    
    
cdef extern from "<math.h>" nogil:
    double fabs(double x)
    double sqrt(double x)
    double atan2(double y, double x)
    double pow(double x, double y)
    
cdef extern from "<stdlib.h>" nogil:
    long abs (long number)
    
    
    
    


##############################################################################
# V1: Implemented all methods to do 2D coupling (monolayer particle coupling)
# using C, both directly and with look up tables. Also C version of particle 
# outgoing-to-outgoing translations
#
# V2: Implemented methods to do 3D coupling. Differentiated between a 2D and 3D
# Legendre and direct coupling functions. Look up tables are NOT available for
# 3D and therefore, explicitly setting the name to 2D was not done. 
#
# V3: Minor changes that I no longer remember... :(
#
# V4: Major changes to all functions such that the Gil is released (at least 
# partially) in all cdef functions. This is good manners, especially for 
# programs like Dask, that check the status of a python program and does not
# know how to interpret when a Cython function holds the Gil then disapears. 
# https://stackoverflow.com/questions/54769782/distributed-worker-warning-heartbeat-to-scheduler-failed
# The release of the Gil also allows multithreading, which is implemented using
# Cython's prange function.
# https://cython.readthedocs.io/en/latest/src/userguide/parallelism.html
# https://docs.dask.org/en/stable/best-practices.html
#
# V5: Minor bug fixes comparing python to cython. Note: AB coeffs are now 
# enforced real/imag. All int -> long
# 
# V6: Proved the ability to call Cython functions from Numba using the "api" command.
# This was shown on the wofz function. 
#
# Cython vs. Python implementations were tested for speed and accuracy. 
# Values output by the Cython funcitons sub-functions match Python sub-functions
# up to the 16 decimal point.
# Cython can give a speed boost from 80-800x, dependent on what functions 
# are being compared and the state of the machine at runtime. 
##############################################################################



'''
##############################################################################
#****************************************************************************#
#****************************************************************************#
#******************** Sub-functions accelerated with C **********************#
#****************************************************************************#
#****************************************************************************#
##############################################################################

This section details all the subfunctions.
'''

##############################################################################
# Calculate wofz special function in a numba-compatible fashion
##############################################################################

# Numba  requires a compatible version of the Faddevva function - replacing scipy.special.wofz().
# The cwrapper is used in smuthi/periodicboundaries/ewald_helper.py and 
# is copied from: https://github.com/numba/numba/issues/3086
# The cython portion of this solution is found below. The purpose is to have a 
# singular file housing all cython functions because this is easier to track/manage.
# This file is automatically compiled for the user and the "api" generates 
# a header file that numba can find. 


cdef api wofz(double in1_real,double in1_imag,double *out_real,double *out_imag):
  cdef double complex z
  z.real=in1_real
  z.imag=in1_imag
  cdef double complex out = scipy.special.cython_special.wofz(z)
  out_real[0]=out.real
  out_imag[0]=out.imag
  return



##############################################################################
# Recursively calculate factorial (n!) - Speed ~2x when called from Python wrap
##############################################################################
cdef double c_factorial(double n) nogil:
    if n == 0:
        return 1
    elif n < 0:
        return 0 #should be NAN!
    else:
        return n * c_factorial(n-1)

def factorial(double n):
    cdef double val
    with nogil:
         val = c_factorial(n)
    return val

##############################################################################
# Recursively calculate double factorial (n!!) - Speed ~1x when called from Python wrap
##############################################################################  
cdef double c_double_factorial(double n) nogil:
    if n in (0, 1):
        return 1
    elif n < 0:
        return 0 #should be NAN!
    else:
        return n * c_double_factorial(n - 2)   

def double_factorial(double n):
    cdef double val
    with nogil:
        val = c_double_factorial(n)
    return val

##############################################################################
# Determine matrix size needed for direct coupling between 2 particles
############################################################################## 
cdef long c_blocksize(long l_max, long m_max) nogil:
    return c_multi_to_single_index(1, l_max, m_max,l_max,m_max) + 1

def blocksize(long l_max, long m_max):
    cdef long val
    with nogil:
        val = c_blocksize(l_max, m_max)
    return val

##############################################################################
# Map tau, l, m to a single linear index - Speed ~0.3x when called from Python wrap
############################################################################## 
cdef long c_multi_to_single_index(long tau,
                                 long l,
                                 long m,
                                 long l_max,
                                 long m_max) nogil:
    cdef long tau_blocksize
    cdef long n
    
    tau_blocksize = m_max * (m_max + 2) + (l_max - m_max) * (2 * m_max + 1)
    n = tau * tau_blocksize
    if (l - 1) <= m_max:
        n += (l - 1) * (l - 1 + 2)
    else:
        n += m_max * (m_max + 2) + (l - 1 - m_max) * (2 * m_max + 1)
    n += m + min(l, m_max)
    return n

def multi_to_single_index(long tau, long l, long m, long l_max, long m_max):
    cdef long val
    with nogil:
        val = c_multi_to_single_index(tau, l, m, l_max, m_max)
    return val

##############################################################################
#  Spherical Hankel function - Speed ~1x when called from Python wrap
############################################################################## 
cdef double complex c_spherical_hankel(long n, double complex x) nogil:
    cdef double complex spherej
    cdef double complex spherey
    cdef double complex sphereh
    
    spherej = scipy.special.cython_special.spherical_jn(n, x)
    if x == 0:
        spherey = -1e24
    else:
        spherey = scipy.special.cython_special.spherical_yn(n, x)
        
    #if spherej < 1e-15:
    #    spherej = 0
    #if spherey < 1e-15:
    #    spherey = 0        
    
    sphereh = spherej + 1j * spherey  
    return sphereh  

def spherical_hankel(long n,double complex x):
    cdef double complex val
    with nogil:
        val = c_spherical_hankel(n,x)
    return val

##############################################################################
#  Spherical Bessel function for translation matrix
############################################################################## 
cdef double complex c_spherical_bessel(long n, double complex x) nogil:
    return scipy.special.cython_special.spherical_jn(n, x)


def spherical_bessel(long n,double complex x):
    cdef double complex val
    with nogil:
        val = c_spherical_bessel(n,x)
    return val

##############################################################################
#  2D Legendre function for direct coupling block - Speed ~153x when called from Python wrap
############################################################################## 
@cython.boundscheck(False)
@cython.wraparound(False)
cdef double[:,:] c_legendre_normalized_2D(long lmax) nogil:
    
    cdef double[:,:] plm
    cdef long l
    cdef long m

    with gil:
        plm = np.zeros((lmax+1,lmax+1),dtype = np.float64)    
    
    plm[0][0] = sqrt(2)/2
    plm[1][0] = 0

    for l in range(1, lmax):
        plm[l + 1][0] = (-l / (l + 1)) * sqrt((2 * l + 3) / (2 * l - 1)) * plm[l-1][0]

    for m in range(1, lmax + 1):
        plm[m][m] = sqrt((2 * m + 1) / (2 * c_factorial(2 * m))) * c_double_factorial(2 * m - 1) 

        for l in range(m, lmax):
            plm[l + 1][m] = -sqrt(((2 * l + 3) * (l - m) * (l + m)) / ((2 * l - 1) * (l + 1 - m) * (l + 1 + m))) * plm[l - 1][m]

    return plm  

def legendre_normalized_2D(long lmax):
    cdef double[:,:] val
    with nogil:
         val = c_legendre_normalized_2D(lmax)
    return np.asarray(val, dtype = np.float64)

##############################################################################
#  3D Legendre function for direct coupling block
############################################################################## 
@cython.boundscheck(False)
@cython.wraparound(False)
cdef double[:,:] c_legendre_normalized_3D(double ct, double st, long lmax) nogil:

    cdef double[:,:] plm
    cdef long l
    cdef long m

    with gil:
        plm = np.zeros((lmax+1,lmax+1),dtype = np.float64)    
    

    plm[0][0] = sqrt(2)/2
    plm[1][0] = sqrt(3/2) * ct

    for l in range(1, lmax):
        plm[l + 1][0] = (1 / (l + 1) * sqrt((2 * l + 1) * (2 * l + 3)) * ct * plm[l][0] -
                         l / (l + 1) * sqrt((2 * l + 3) / (2 * l - 1)) * plm[l-1][0])
    for m in range(1, lmax + 1):
        plm[m][m] = sqrt((2 * m + 1) / (2 * c_factorial(2 * m))) * c_double_factorial(2 * m - 1) * st**m
        for l in range(m, lmax):
            plm[l + 1][m] = (sqrt((2 * l + 1) * (2 * l + 3) / ((l + 1 - m) * (l + 1 + m))) * ct * plm[l][m] -
                             sqrt((2 * l + 3) * (l - m) * (l + m) / ((2 * l - 1) * (l + 1 - m) * (l + 1 + m))) *
                             plm[l - 1][m])
    return plm  

def legendre_normalized_3D(double ct, double st, long lmax):
    cdef double[:,:] val
    with nogil:
        val = c_legendre_normalized_3D(ct, st, lmax)
    return np.asarray(val)


##############################################################################
#  AB5 coefficients - Speed ~2x when called from Python wrap
############################################################################## 
from numba.core.typing import cffi_utils
from pywigxjpf_ffi import ffi, lib
import pywigxjpf_ffi
cffi_utils.register_module(pywigxjpf_ffi)
nb_wig3jj = pywigxjpf_ffi.lib.wig3jj

lib.wig_table_init(100,9)
lib.wig_temp_init(100)


@cython.boundscheck(False)
@cython.wraparound(False)
cdef (double complex, double complex) c_ab5_coefficients(long l1,
                        long m1,
                        long l2, 
                        long m2, 
                        long p) nogil:
    
    cdef double complex jfac = 0+1j*0
    cdef double complex fac1 = 0+1j*0
    cdef double complex fac2a = 0+1j*0
    cdef double complex fac2b = 0+1j*0
    cdef double complex a ,b
    cdef double wig1, wig2a, wig2b,

    jfac = cpow(1.0j,(abs(m1 - m2) - abs(m1) - abs(m2) + l2 - l1 + p)) * (cpow((-1), abs((m1 - m2))))
    fac1 = sqrt(((2 * l1 + 1) * (2 * l2 + 1)) / (2 * l1 * (l1 + 1) * l2 * (l2 + 1)))
    fac2a = (l1 * (l1 + 1) + l2 * (l2 + 1) - p * (p + 1)) * sqrt(2 * p + 1)
    fac2b = sqrt((l1 + l2 + 1 + p) * (l1 + l2 + 1 - p) * (p + l1 - l2) * (p - l1 + l2) * (2 * p + 1))
    
    # Note that arguments are in two_j = 2*j
    with gil:
        wig1 = nb_wig3jj(2*l1, 2*l2, 2*p, 2*m1, -m2*2, -(m1 - m2)*2)
        wig2a = nb_wig3jj(2*l1, 2*l2, 2*p, 0, 0, 0)
        wig2b = nb_wig3jj(2*l1, 2*l2, 2*(p - 1), 0, 0, 0)

    # Instead of 0 the result gives numbers <e-15. Force these to 0 so it matches python.
    # The problem is probably in the jfac 
    a = creal(jfac * fac1 * fac2a * wig1 * wig2a)+1j*0  # As are always real.
    b = 0+1j*cimag(jfac * fac1 * fac2b * wig1 * wig2b)  # Bs are always imag.
    
    return a, b

def cab5_coefficients(long l1, long m1, long l2, long m2, long p):
    cdef double complex a
    cdef double complex b
    with nogil:
        a,b = c_ab5_coefficients(l1, m1, l2, m2, p)
    return complex(a), complex(b)


'''
##############################################################################
#****************************************************************************#
#****************************************************************************#
#**************** 3D direct coupling block and translation ******************#
#********************* Includes multithread options *************************#
#****************************************************************************#
##############################################################################

This section details the direct coupling block and translation between 2 
particles that can be anywhere in the same layer.

Since the direct coupling block is actually just a translation using a different
spherical bessel funciton, use a "kind" flag to chose the bessel function. Then, 
implement python wrappers to call this flag appropriately based on implementation.
This prevents coding error and makes the code more readable.

Options for multithreading are added. 
In case these options cause errors for different implementations/configurations, 
multithreading compatible functions are given a different name so a fall 
back option is accessable.
'''

##############################################################################
# Translate block between 2 particles in 3D
##############################################################################
#@cython.cdivision(True) <- DO NOT ENABLE THIS. IT WILL DO INTERGER DIVISION!
@cython.boundscheck(False)
@cython.wraparound(False)
cdef double complex [:,:] c_svwf_translate_3D(double complex k,
                                                  double dx,
                                                  double dy,
                                                  double dz,
                                                  long lmax1,
                                                  long lmax2,
                                                  long mmax1,
                                                  long mmax2, 
                                                  long kind) nogil:

    # Type things and calculate simple 1 line formulas
    # Note: Math functions are optimized by standard c libraries
    cdef long blocksize1 = c_blocksize(lmax1, mmax1)
    cdef long blocksize2 = c_blocksize(lmax2, mmax2)
    cdef double complex d = csqrt(pow(dx,2) + pow(dy,2) + pow(dz,2))
    cdef double d_real = sqrt(pow(dx,2) + pow(dy,2) + pow(dz,2))
    cdef double cos_theta = dz / d_real
    cdef double sin_theta = sqrt(dx**2 + dy**2) / d_real
    cdef double complex kd = k*d
    cdef double[:,:] legendre
    cdef double complex[:,::1] w 
    cdef double complex [:] sph
    cdef int[:] m1_array, m2_array
    cdef long im1, im2, l1, l2, ld, n1, n2, tau1, tau2, i, m1, m2
    cdef double complex A, B, a5, b5, eimph
    cdef double phi 
    
    # All numpy initialize functions must be done with the Gil. 
    # Other methods such as malloc or pointers can be used for pure C
    # Though, I see no reason. 
    with gil:
        phi = np.arctan2(dy, dx)
        w = np.zeros((blocksize1, blocksize2), dtype=np.complex128, order = 'C')   
        # If the particle is itself then coupling matrix = 0 matrix.
        if d == 0:
            return w
        sph = np.zeros((lmax1+lmax2+1), dtype=np.complex128)
        m1_array = np.arange(-mmax1, mmax1+1, 1, dtype = np.intc)
        m2_array = np.arange(-mmax2, mmax2+1, 1, dtype = np.intc)

    if kind == 0: # 0 = incoming wave, 1 = outgoing wave. 
    #NOTE: an long flag is used instead of a string because Cython 
    # has troublesome conventions about datatyping strings from python v2 to v3
    # See more on this at https://cython.readthedocs.io/en/latest/src/tutorial/strings.html
        for n in range(lmax1 + lmax2 + 1):
            sph[n] = c_spherical_hankel(n, kd) 
    else:
        for n in range(lmax1 + lmax2 + 1):
            sph[n] = c_spherical_bessel(n, kd)       
        
    legendre = c_legendre_normalized_3D(cos_theta, sin_theta, lmax1 + lmax2) 

    
    for im1 in range(0, 2*mmax1 + 1):
        m1 = m1_array[im1]
        for im2 in range(0, 2*mmax2 + 1):
            m2 = m2_array[im2]
            eimph = cexp(1j * (m2 - m1) * phi)
            for l1 in range(max(1, abs(m1)), lmax1 + 1):
                for l2 in range(max(1, abs(m2)), lmax2 + 1):
                    A = 0
                    B = 0
                    for ld in range(max(abs(l1 - l2), abs(m1 - m2)), l1 + l2 + 1):  
                        a5, b5 = c_ab5_coefficients(l2, m2, l1, m1, ld)
                        A += a5 * sph[ld] * legendre[ld][abs(m1 - m2)]
                        B += b5 * sph[ld] * legendre[ld][abs(m1 - m2)]
                    A, B = eimph * A, eimph * B
                    for tau1 in range(2):
                        n1 = c_multi_to_single_index(tau1, l1, m1, lmax1, mmax1)
                        for tau2 in range(2):
                            n2 = c_multi_to_single_index(tau2, l2, m2, lmax2, mmax2)
                            if tau1 == tau2:
                                w[n1, n2] = A
                            else:
                                w[n1, n2] = B

    return w

##############################################################################
# Translate block between 2 particles in 3D with multithreading enabled
##############################################################################
@cython.boundscheck(False)
@cython.wraparound(False)
cdef double complex [:,:] c_svwf_translate_3D_threaded(double complex k,
                                                  double dx,
                                                  double dy,
                                                  double dz,
                                                  long lmax1,
                                                  long lmax2,
                                                  long mmax1,
                                                  long mmax2, 
                                                  long kind) nogil:

    # Type things and calculate simple 1 line formulas
    # Note: Math functions are optimized by standard c libraries
    cdef long blocksize1 = c_blocksize(lmax1, mmax1)
    cdef long blocksize2 = c_blocksize(lmax2, mmax2)
    cdef double complex d = csqrt(pow(dx,2) + pow(dy,2) + pow(dz,2))
    cdef double d_real = sqrt(pow(dx,2) + pow(dy,2) + pow(dz,2))
    cdef double cos_theta = dz / d_real
    cdef double sin_theta = sqrt(dx**2 + dy**2) / d_real
    cdef double complex kd = k*d
    cdef double[:,:] legendre
    cdef double complex[:,::1] w 
    cdef double complex [:] sph
    cdef int[:] m1_array, m2_array
    cdef long im1, im2, l1, l2, ld, n1, n2, tau1, tau2, i, m1, m2
    cdef double complex A, B, a5, b5, eimph
    cdef double phi 
    
    # All numpy initialize functions must be done with the Gil. 
    # Other methods such as malloc or pointers can be used for pure C
    # Though, I see no reason. 
    with gil:
        phi = np.arctan2(dy, dx)
        w = np.zeros((blocksize1, blocksize2), dtype=np.complex128, order = 'C')   
        # If the particle is itself then coupling matrix = 0 matrix.
        if d == 0:
            return w
        sph = np.zeros((lmax1+lmax2+1), dtype=np.complex128)
        m1_array = np.arange(-mmax1, mmax1+1, 1, dtype = np.intc)
        m2_array = np.arange(-mmax2, mmax2+1, 1, dtype = np.intc)

    if kind == 0: # 0 = incoming wave, 1 = outgoing wave. 
    #NOTE: an long flag is used instead of a string because Cython 
    # has troublesome conventions about datatyping strings from python v2 to v3
    # See more on this at https://cython.readthedocs.io/en/latest/src/tutorial/strings.html
        for n in range(lmax1 + lmax2 + 1):
            sph[n] = c_spherical_hankel(n, kd) 
    else:
        for n in range(lmax1 + lmax2 + 1):
            sph[n] = c_spherical_bessel(n, kd)       
        
    legendre = c_legendre_normalized_3D(cos_theta, sin_theta, lmax1 + lmax2) 

    
    for im1 in prange(0, 2*mmax1 + 1): #<- only change to thread since gil already released!
        m1 = m1_array[im1]
        for im2 in range(0, 2*mmax2 + 1):
            m2 = m2_array[im2]
            eimph = cexp(1j * (m2 - m1) * phi)
            for l1 in range(max(1, abs(m1)), lmax1 + 1):
                for l2 in range(max(1, abs(m2)), lmax2 + 1):
                    A = 0
                    B = 0
                    for ld in range(max(abs(l1 - l2), abs(m1 - m2)), l1 + l2 + 1):  
                        a5, b5 = c_ab5_coefficients(l2, m2, l1, m1, ld)
                        A += a5 * sph[ld] * legendre[ld][abs(m1 - m2)]
                        B += b5 * sph[ld] * legendre[ld][abs(m1 - m2)]
                    A, B = eimph * A, eimph * B
                    for tau1 in range(2):
                        n1 = c_multi_to_single_index(tau1, l1, m1, lmax1, mmax1)
                        for tau2 in range(2):
                            n2 = c_multi_to_single_index(tau2, l2, m2, lmax2, mmax2)
                            if tau1 == tau2:
                                w[n1, n2] = A
                            else:
                                w[n1, n2] = B

    return w


# Have option to thread build into the translate function for easier reuse.
def svwf_translate_3D(double complex k, double dx, double dy, double dz,
                      long lmax1, long lmax2, long mmax1, long mmax2,
                      long kind, threaded = False):
    cdef double complex [:,:] val
    
    if threaded == True:
        with nogil:
            val = c_svwf_translate_3D_threaded(k, dx, dy, dz, lmax1, lmax2, mmax1, mmax2, kind)
    else:
        with nogil:
            val = c_svwf_translate_3D(k, dx, dy, dz, lmax1, lmax2, mmax1, mmax2, kind)
    return np.asarray(val, dtype = np.complex128)

# Special case of translate. Make wrapper for readability in smuthi.
def direct_coupling_block_3D(double complex k, double dx, double dy, double dz,
                              long lmax1, long lmax2, long mmax1, long mmax2, threaded = False):
    cdef kind = 0
    return svwf_translate_3D(k, dx, dy,  dz,
                             lmax1,lmax2,mmax1,mmax2,
                             kind, threaded)


'''
##############################################################################
#****************************************************************************#
#****************************************************************************#
#********** 3D direct coupling block and translation hash   *****************#
#********************* Includes multithread options *************************#
#****************************************************************************#
##############################################################################

The same AB5 coefficients are often shared between particle pairs. In Python
these are memoized. In Cython, we memoize a hash table of particle order 
pairs (e.g., 2->2 , 3->2, 2->3, ...). This table is then used for fast C loops.

This section details the direct coupling block and translation between 2 
particles that can be anywhere in the same layer. Options for multithreading 
are added. In case these options cause errors for different implementations, 
multithreading compatible functions are given a different name so a fall 
back option is accessable.
'''

##############################################################################
#  Make  hash table of ab5 for a particle order pair (WITH GIL)
##############################################################################
# """
# This hash table is particularly good when all particles have the same order. 
# """
@cython.boundscheck(False)
@cython.wraparound(False)
cdef c_hard_ab5_coefficient_hash_table(long lmax1, 
                                 long lmax2, 
                                 long mmax1, 
                                 long mmax2):   #<-- Note: This function holds the gil!
    
    cdef long ab5counter = 0
    cdef long m1, m2, l1, l2, ld
    for m1 in range(-mmax1, mmax1 + 1):
        for m2 in range(-mmax2, mmax2 + 1):
            for l1 in range(max(1, abs(m1)), lmax1 + 1):
                for l2 in range(max(1, abs(m2)), lmax2 + 1):
                    for ld in range(max(abs(l1 - l2), abs(m1 - m2)), l1 + l2 + 1): 
                        ab5counter += 1
    
   
    cdef double complex[:] a5_array, b5_array
    cdef double complex a5, b5
    
    a5_array = np.ones((ab5counter,), dtype=np.complex128)
    b5_array = np.ones((ab5counter,), dtype=np.complex128)
   
    ab5counter = 0
    for m1 in range(-mmax1, mmax1 + 1):
        for m2 in range(-mmax2, mmax2 + 1):
            for l1 in range(max(1, abs(m1)), lmax1 + 1):
                for l2 in range(max(1, abs(m2)), lmax2 + 1):
                    for ld in range(max(abs(l1 - l2), abs(m1 - m2)), l1 + l2 + 1): 
                        a5, b5 = c_ab5_coefficients(l2, m2, l1, m1, ld)
                        a5_array[ab5counter] = a5
                        b5_array[ab5counter] = b5
                        ab5counter+=1
                        
    return a5_array, b5_array

def _ab5_coefficient_hash_table(long lmax1, long lmax2, long mmax1, long mmax2):
    cdef double complex[:] a5_array, b5_array
    a5_array, b5_array = c_hard_ab5_coefficient_hash_table(lmax1, lmax2, mmax1, mmax2)
    return np.asarray(a5_array, dtype = np.complex128),  np.asarray(b5_array, dtype = np.complex128)



##############################################################################
# Translation 3D with  Hash
##############################################################################
@cython.boundscheck(False)
@cython.wraparound(False)
cdef double complex [:,:]  c_svwf_translate_3D_from_hash_table(long blocksize1,long blocksize2,
                            double complex[:,:] w,
                            double complex[:] sph,
                            double complex k,
                            double dx, double dy, double dz,
                            long lmax1, long lmax2, long mmax1, long mmax2,
                            double complex[:] a5_array, 
                            double complex[:] b5_array, 
                            long kind) nogil:



    # calculate distance vector
    cdef double complex d = csqrt(pow(dx,2) + pow(dy,2) + pow(dz,2))
    cdef double d_real = sqrt(pow(dx,2) + pow(dy,2) + pow(dz,2))
    cdef double cos_theta = dz / d_real
    cdef double sin_theta = sqrt(dx**2 + dy**2) / d_real
    cdef double complex kd = k*d
    cdef double[:,:] legendre
    cdef int[:] m1_array, m2_array
    
    # calculate coupling matrix
    cdef long im1, im2, l1, l2, ld, n1, n2, tau1, tau2, i, m1, m2
    cdef double complex A, B, a5, b5, eimph
    cdef double phi = atan2(dy, dx)
    
    with gil:
        # If the particle is itself then coupling matrix = 0 matrix.
        m1_array = np.arange(-mmax1, mmax1+1, 1, dtype = np.intc)
        m2_array = np.arange(-mmax2, mmax2+1, 1, dtype = np.intc)
        
        
    if d == 0:
        return w    

    if kind == 0: # 0 = incoming wave, 1 = outgoing wave. 
    #NOTE: an long flag is used instead of a string because Cython 
    # has troublesome conventions about datatyping strings from python v2 to v3
    # See more on this at https://cython.readthedocs.io/en/latest/src/tutorial/strings.html
        for n in range(lmax1 + lmax2 + 1):
            sph[n] = c_spherical_hankel(n, kd) 
    else:
        for n in range(lmax1 + lmax2 + 1):
            sph[n] = c_spherical_bessel(n, kd)      
    legendre = c_legendre_normalized_3D(cos_theta, sin_theta, lmax1 + lmax2) 

    cdef long ab5counter = 0
    for im1 in range(0, 2*mmax1 + 1): # <- Only necessary change for threading since gil is already released 
        m1 = m1_array[im1]
        for im2 in range(0, 2*mmax2 + 1):
            m2 = m2_array[im2]
            eimph = cexp(1j * (m2 - m1) * phi)
            for l1 in range(max(1, abs(m1)), lmax1 + 1):
                for l2 in range(max(1, abs(m2)), lmax2 + 1):
                    A = 0
                    B = 0
                    for ld in range(max(abs(l1 - l2), abs(m1 - m2)), l1 + l2 + 1):  
                        A += a5_array[ab5counter] * sph[ld] * legendre[ld][abs(m1 - m2)]
                        B += b5_array[ab5counter] * sph[ld] * legendre[ld][abs(m1 - m2)]
                        ab5counter+=1
                    A, B = eimph * A, eimph * B
                    for tau1 in range(2):
                        n1 = c_multi_to_single_index(tau1, l1, m1, lmax1, mmax1)
                        for tau2 in range(2):
                            n2 = c_multi_to_single_index(tau2, l2, m2, lmax2, mmax2)
                            if tau1 == tau2:
                                w[n1, n2] = A
                            else:
                                w[n1, n2] = B

    return w


##############################################################################
# Translation 3D with  Hash and Multithreading
##############################################################################
@cython.boundscheck(False)
@cython.wraparound(False)
cdef double complex [:,:]  c_svwf_translate_3D_from_hash_table_threaded(long blocksize1,long blocksize2,
                            double complex[:,:] w,
                            double complex[:] sph,
                            double complex k,
                            double dx, double dy, double dz,
                            long lmax1, long lmax2, long mmax1, long mmax2,
                            double complex[:] a5_array, 
                            double complex[:] b5_array,
                            long kind) nogil:



    # calculate distance vector
    cdef double complex d = csqrt(pow(dx,2) + pow(dy,2) + pow(dz,2))
    cdef double d_real = sqrt(pow(dx,2) + pow(dy,2) + pow(dz,2))
    cdef double cos_theta = dz / d_real
    cdef double sin_theta = sqrt(dx**2 + dy**2) / d_real
    cdef double complex kd = k*d
    cdef double[:,:] legendre
    cdef int[:] m1_array, m2_array
    
    # calculate coupling matrix
    cdef long im1, im2, l1, l2, ld, n1, n2, tau1, tau2, i, m1, m2
    cdef double complex A, B, a5, b5, eimph
    cdef double phi = atan2(dy, dx)
    
    with gil:
        # If the particle is itself then coupling matrix = 0 matrix.
        m1_array = np.arange(-mmax1, mmax1+1, 1, dtype = np.intc)
        m2_array = np.arange(-mmax2, mmax2+1, 1, dtype = np.intc)
        
        
    if d == 0:
        return w    

    if kind == 0: # 0 = incoming wave, 1 = outgoing wave. 
    #NOTE: an long flag is used instead of a string because Cython 
    # has troublesome conventions about datatyping strings from python v2 to v3
    # See more on this at https://cython.readthedocs.io/en/latest/src/tutorial/strings.html
        for n in range(lmax1 + lmax2 + 1):
            sph[n] = c_spherical_hankel(n, kd) 
    else:
        for n in range(lmax1 + lmax2 + 1):
            sph[n] = c_spherical_bessel(n, kd)      
    legendre = c_legendre_normalized_3D(cos_theta, sin_theta, lmax1 + lmax2) 

    cdef long ab5counter = 0
    for im1 in prange(0, 2*mmax1 + 1): # <- Only necessary change for threading since gil is already released 
        m1 = m1_array[im1]
        for im2 in range(0, 2*mmax2 + 1):
            m2 = m2_array[im2]
            eimph = cexp(1j * (m2 - m1) * phi)
            for l1 in range(max(1, abs(m1)), lmax1 + 1):
                for l2 in range(max(1, abs(m2)), lmax2 + 1):
                    A = 0
                    B = 0
                    for ld in range(max(abs(l1 - l2), abs(m1 - m2)), l1 + l2 + 1):  
                        A += a5_array[ab5counter] * sph[ld] * legendre[ld][abs(m1 - m2)]
                        B += b5_array[ab5counter] * sph[ld] * legendre[ld][abs(m1 - m2)]
                        ab5counter+=1
                    A, B = eimph * A, eimph * B
                    for tau1 in range(2):
                        n1 = c_multi_to_single_index(tau1, l1, m1, lmax1, mmax1)
                        for tau2 in range(2):
                            n2 = c_multi_to_single_index(tau2, l2, m2, lmax2, mmax2)
                            if tau1 == tau2:
                                w[n1, n2] = A
                            else:
                                w[n1, n2] = B

    return w


# Have option to thread build into the translate function for easier reuse.
def svwf_translate_3D_from_hash_table(long blocksize1,long blocksize2,
                            double complex[:,:] w,
                            double complex[:] sph,
                            double complex k,
                            double dx, double dy, double dz,
                            long lmax1, long lmax2, long mmax1, long mmax2,
                            double complex[:] a5_array, 
                            double complex[:] b5_array,
                            long kind, threaded = False):
    
    cdef double complex [:,:] val
    if threaded == True:
        with nogil:
            val = c_svwf_translate_3D_from_hash_table_threaded(blocksize1,blocksize2,
                                            w,sph,k, dx, dy,  dz,
                                            lmax1,lmax2,mmax1,mmax2,
                                            a5_array, b5_array,
                                            kind)
    else:
        with nogil:
            val = c_svwf_translate_3D_from_hash_table(blocksize1,blocksize2,
                                            w,sph,k, dx, dy,  dz,
                                            lmax1,lmax2,mmax1,mmax2,
                                            a5_array, b5_array,
                                            kind)
            
    return np.asarray(val, dtype = np.complex128)

# Special case of translate. Make wrapper for readability in smuthi.
def direct_coupling_block_3D_from_hash_table(long blocksize1,long blocksize2,
                                double complex[:,:] w,
                                double complex[:] sph,
                                double complex k,
                                double dx, double dy, double dz,
                                long lmax1, long lmax2, long mmax1, long mmax2,
                                double complex[:] a5_array, 
                                double complex[:] b5_array,                             
                                threaded = False):
    cdef kind = 0
    return svwf_translate_3D_from_hash_table(blocksize1,blocksize2,
                                    w,sph,k, dx, dy,  dz,
                                    lmax1,lmax2,mmax1,mmax2,
                                    a5_array, b5_array,
                                    kind, threaded)


'''
##############################################################################
#****************************************************************************#
#****************************************************************************#
#**************** 2D direct coupling block and translation ******************#
#********************* Includes multithread options *************************#
#****************************************************************************#
##############################################################################

This section details the direct coupling block and translation between 2 
particles that can be anywhere in the same x-y plane. Options for multithreading 
are added. In case these options cause errors for different implementations, 
multithreading compatible functions are given a different name so a fall 
back option is accessable.
'''


##############################################################################
# Translate block between 2 particles in 2D (Legendre = 90 deg)
##############################################################################


@cython.boundscheck(False)
@cython.wraparound(False)
cdef double complex [:,:] c_svwf_translate_2D(double complex k,
                                                  double dx,
                                                  double dy,
                                                  double dz,
                                                  long lmax1,
                                                  long lmax2,
                                                  long mmax1,
                                                  long mmax2,
                                                  long kind) nogil:

    # Get index
    cdef long blocksize1 = c_blocksize(lmax1, mmax1)
    cdef long blocksize2 = c_blocksize(lmax2, mmax2)
    
    # calculate coupling matrix
    cdef long im1, im2, l1, l2, ld, n1, n2, tau1, tau2, i, m1, m2
    cdef double complex A, B, a5, b5, eimph, d, phi, kd
    cdef double complex[:,::1] w
    cdef double complex [:] sph
    cdef int[:] m1_array, m2_array 
    cdef double[:,:] legendre
    
    with gil:
        # calculate distance vector
        d = sqrt(pow(dx,2) + pow(dy,2) + pow(dz,2))
        phi = np.arctan2(dy, dx)

        #initialize coupling matrix
        w = np.zeros((blocksize1, blocksize2), dtype=np.complex128, order = 'C')
        # If the particle is itself then coupling matrix = 0 matrix.
        if d == 0:
            return w
        sph = np.zeros((lmax1+lmax2+1), dtype=np.complex128)
        m1_array = np.arange(-mmax1, mmax1+1, 1, dtype = np.intc)
        m2_array = np.arange(-mmax2, mmax2+1, 1, dtype = np.intc)



    # spherical functions  
    kd = k*d
    if kind == 0: # 0 = incoming wave, 1 = outgoing wave. 
    #NOTE: an long flag is used instead of a string because Cython 
    # has troublesome conventions about datatyping strings from python v2 to v3
    # See more on this at https://cython.readthedocs.io/en/latest/src/tutorial/strings.html
        for n in range(lmax1 + lmax2 + 1):
            sph[n] = c_spherical_hankel(n, kd) 
    else:
        for n in range(lmax1 + lmax2 + 1):
            sph[n] = c_spherical_bessel(n, kd)  
    legendre = c_legendre_normalized_2D(lmax1 + lmax2) 

    
    for im1 in range(0, 2*mmax1 + 1):
        m1 = m1_array[im1]
        for im2 in range(0, 2*mmax2 + 1):
            m2 = m2_array[im2]
            eimph = cexp(1j * (m2 - m1) * phi)
            for l1 in range(max(1, abs(m1)), lmax1 + 1):
                for l2 in range(max(1, abs(m2)), lmax2 + 1):
                    A = 0
                    B = 0
                    for ld in range(max(abs(l1 - l2), abs(m1 - m2)), l1 + l2 + 1):  
                        a5, b5 = c_ab5_coefficients(l2, m2, l1, m1, ld)
                        A += a5 * sph[ld] * legendre[ld][abs(m1 - m2)]
                        B += b5 * sph[ld] * legendre[ld][abs(m1 - m2)]
                    A, B = eimph * A, eimph * B
                    for tau1 in range(2):
                        n1 = c_multi_to_single_index(tau1, l1, m1, lmax1, mmax1)
                        for tau2 in range(2):
                            n2 = c_multi_to_single_index(tau2, l2, m2, lmax2, mmax2)
                            if tau1 == tau2:
                                w[n1, n2] = A
                            else:
                                w[n1, n2] = B

    return w



##############################################################################
# Translate block between 2 particles in 2D (Legendre = 90 deg) with threading
##############################################################################
@cython.boundscheck(False)
@cython.wraparound(False)
cdef double complex [:,:] c_svwf_translate_2D_threaded(double complex k,
                                                  double dx,
                                                  double dy,
                                                  double dz,
                                                  long lmax1,
                                                  long lmax2,
                                                  long mmax1,
                                                  long mmax2,
                                                  long kind) nogil:

    # Get index
    cdef long blocksize1 = c_blocksize(lmax1, mmax1)
    cdef long blocksize2 = c_blocksize(lmax2, mmax2)
    
    # calculate coupling matrix
    cdef long im1, im2, l1, l2, ld, n1, n2, tau1, tau2, i, m1, m2
    cdef double complex A, B, a5, b5, eimph, d, phi, kd
    cdef double complex[:,::1] w
    cdef double complex [:] sph
    cdef int[:] m1_array, m2_array 
    cdef double[:,:] legendre
    
    with gil:
        # calculate distance vector
        d = sqrt(pow(dx,2) + pow(dy,2) + pow(dz,2))
        phi = np.arctan2(dy, dx)

        #initialize coupling matrix
        w = np.zeros((blocksize1, blocksize2), dtype=np.complex128, order = 'C')
        # If the particle is itself then coupling matrix = 0 matrix.
        if d == 0:
            return w
        sph = np.zeros((lmax1+lmax2+1), dtype=np.complex128)
        m1_array = np.arange(-mmax1, mmax1+1, 1, dtype = np.intc)
        m2_array = np.arange(-mmax2, mmax2+1, 1, dtype = np.intc)



    # spherical functions  
    kd = k*d
    if kind == 0: # 0 = incoming wave, 1 = outgoing wave. 
    #NOTE: an long flag is used instead of a string because Cython 
    # has troublesome conventions about datatyping strings from python v2 to v3
    # See more on this at https://cython.readthedocs.io/en/latest/src/tutorial/strings.html
        for n in range(lmax1 + lmax2 + 1):
            sph[n] = c_spherical_hankel(n, kd) 
    else:
        for n in range(lmax1 + lmax2 + 1):
            sph[n] = c_spherical_bessel(n, kd)  
    legendre = c_legendre_normalized_2D(lmax1 + lmax2) 

    
    for im1 in prange(0, 2*mmax1 + 1):
        m1 = m1_array[im1]
        for im2 in range(0, 2*mmax2 + 1):
            m2 = m2_array[im2]
            eimph = cexp(1j * (m2 - m1) * phi)
            for l1 in range(max(1, abs(m1)), lmax1 + 1):
                for l2 in range(max(1, abs(m2)), lmax2 + 1):
                    A = 0
                    B = 0
                    for ld in range(max(abs(l1 - l2), abs(m1 - m2)), l1 + l2 + 1):  
                        a5, b5 = c_ab5_coefficients(l2, m2, l1, m1, ld)
                        A += a5 * sph[ld] * legendre[ld][abs(m1 - m2)]
                        B += b5 * sph[ld] * legendre[ld][abs(m1 - m2)]
                    A, B = eimph * A, eimph * B
                    for tau1 in range(2):
                        n1 = c_multi_to_single_index(tau1, l1, m1, lmax1, mmax1)
                        for tau2 in range(2):
                            n2 = c_multi_to_single_index(tau2, l2, m2, lmax2, mmax2)
                            if tau1 == tau2:
                                w[n1, n2] = A
                            else:
                                w[n1, n2] = B

    return w


# Have option to thread build into the translate function for easier reuse.
def svwf_translate_2D(double complex k, double dx, double dy, double dz,
                      long lmax1, long lmax2, long mmax1, long mmax2,
                      long kind, threaded = False):
    
    cdef double complex [:,:] val

    if threaded == True:
        with nogil:
            val = c_svwf_translate_2D_threaded(k, dx, dy, dz, 
                                              lmax1, lmax2, mmax1, mmax2,
                                              kind)
    else:
        with nogil:
            val = c_svwf_translate_2D(k, dx, dy, dz, 
                                     lmax1, lmax2, mmax1, mmax2,
                                     kind)
            
    return np.asarray(val, dtype = np.complex128)

# Special case of translate. Make wrapper for readability in smuthi.
def direct_coupling_block_2D(double complex k, double dx, double dy, double dz,
                              long lmax1, long lmax2, long mmax1, long mmax2,                          
                              threaded = False):
    cdef kind = 0
    return svwf_translate_2D(k, dx, dy, dz,
                             lmax1, lmax2, mmax1, mmax2, 
                             kind, threaded)

'''
##############################################################################
#****************************************************************************#
#****************************************************************************#
#********** 2D direct coupling block and translation hash   *****************#
#********************* Includes multithread options *************************#
#****************************************************************************#
##############################################################################

The same AB5 coefficients are often shared between particle pairs. In Python
these are memoized. In Cython, we memoize a hash table of particle order 
pairs (e.g., 2->2 , 3->2, 2->3, ...). This table is then used for fast C loops.

This section details the direct coupling block and translation between 2 
particles that can be anywhere in the same layer. Options for multithreading 
are added. In case these options cause errors for different implementations, 
multithreading compatible functions are given a different name so a fall 
back option is accessable.
'''



##############################################################################
#  Make  hash table of ab5 and legendre for a particle pair
##############################################################################
# """
#     This hash table is considered "hard" because it is dependent on the order 
#     of which particle is the emitter and which is the reciever. This hash 
#     table is particularly good when all particles have the same order. 
#     If orders are different, a hash of just a5 and b5 as multidimensional 
#     arrays may be a better option because you dont have to run the hash for 
#     e.g., 2->3 and 3->2. But, no tests have been run to verify there is truly 
#     a speed increase... so, unless this version of a hash becomes the 
#     bottleneck I DO NOT recomend writing and debugging a new more "generalized"
#     hash..
# """
@cython.boundscheck(False)
@cython.wraparound(False)
cdef c_hard_ab5_coefficient_and_legendre_hash_table(long lmax1, 
                                             long lmax2, 
                                             long mmax1, 
                                             long mmax2):
    
    cdef long ab5counter = 0
    cdef long m1, m2, l1, l2, ld
    for m1 in range(-mmax1, mmax1 + 1):
        for m2 in range(-mmax2, mmax2 + 1):
            for l1 in range(max(1, abs(m1)), lmax1 + 1):
                for l2 in range(max(1, abs(m2)), lmax2 + 1):
                    for ld in range(max(abs(l1 - l2), abs(m1 - m2)), l1 + l2 + 1): 
                        ab5counter += 1
    
    cdef double complex[:] a5leg_array, b5leg_array
    cdef double complex a5, b5, leg
    cdef double [:,:] legendre

    a5leg_array = np.ones((ab5counter,), dtype=np.complex128)
    b5leg_array = np.ones((ab5counter,), dtype=np.complex128)

    ab5counter = 0
    legendre = c_legendre_normalized_2D(lmax1+lmax2+1)
    for m1 in range(-mmax1, mmax1 + 1):
        for m2 in range(-mmax2, mmax2 + 1):
            for l1 in range(max(1, abs(m1)), lmax1 + 1):
                for l2 in range(max(1, abs(m2)), lmax2 + 1):
                    for ld in range(max(abs(l1 - l2), abs(m1 - m2)), l1 + l2 + 1): 
                        a5, b5 = c_ab5_coefficients(l2, m2, l1, m1, ld)
                        leg =  complex(legendre[ld][abs(m1 - m2)])
                        a5leg_array[ab5counter] = leg*a5
                        b5leg_array[ab5counter] = leg*b5
                        ab5counter+=1
                        
    return a5leg_array, b5leg_array

def _ab5_coefficient_and_legendre_hash_table(long lmax1, long lmax2, long mmax1, long mmax2):
    cdef double complex[:] a5leg_array, b5leg_array 
    a5leg_array, b5leg_array = c_hard_ab5_coefficient_and_legendre_hash_table(lmax1, lmax2, mmax1, mmax2)
    return np.asarray(a5leg_array, dtype = np.complex128),  np.asarray(b5leg_array, dtype = np.complex128)

##############################################################################
#  Direct coupling block using  hash table
##############################################################################
@cython.boundscheck(False)
@cython.wraparound(False)
cdef double complex[:,:] c_svwf_translate_2D_from_hash_table(long blocksize1,
                                                            long blocksize2,
                                                            double complex[:,:] w,
                                                            double complex[:] sph,
                                                            double complex k,
                                                            double dx,
                                                            double dy,
                                                            double dz,
                                                            long lmax1,
                                                            long lmax2,
                                                            long mmax1,
                                                            long mmax2,
                                                            double complex[:] a5leg_hash_table,
                                                            double complex[:] b5leg_hash_table,
                                                            long kind) nogil:      

    cdef double phi = atan2(dy, dx)
    cdef double complex d = csqrt(pow(dx,2) + pow(dy,2) + pow(dz,2))

    # If the particle is itself then coupling matrix = 0 matrix.
    if d == 0:
        return w
    
    # spherical functions
    cdef double complex kd = k*d
    
    if kind == 0: # 0 = incoming wave, 1 = outgoing wave. 
    #NOTE: an long flag is used instead of a string because Cython 
    # has troublesome conventions about datatyping strings from python v2 to v3
    # See more on this at https://cython.readthedocs.io/en/latest/src/tutorial/strings.html
        for n in range(lmax1 + lmax2 + 1):
            sph[n] = c_spherical_hankel(n, kd) 
    else:
        for n in range(lmax1 + lmax2 + 1):
            sph[n] = c_spherical_bessel(n, kd)  
  

    # calculate coupling matrix
    cdef long m1, m2, l1, l2, ld, n1, n2, tau1, tau2, idx_ab5leg, idx_eim 
    cdef double complex A, B, eimph
                                                        
    # the particle coupling operator is the transpose of the SVWF translation operator
    # therefore, (l1,m1) and (l2,m2) are interchanged:
    idx_ab5leg = 0       
    for m1 in range(-mmax1, mmax1 + 1):
        for m2 in range(-mmax2, mmax2 + 1):
            eimph = cexp(1j * (m2 - m1) * phi) 
            for l1 in range(max(1, abs(m1)), lmax1 + 1):
                for l2 in range(max(1, abs(m2)), lmax2 + 1):
                    A = 0
                    B = 0
                    for ld in range(max(abs(l1 - l2), abs(m1 - m2)), l1 + l2 + 1):  # if ld<abs(m1-m2) then P=0
                       
                        A += a5leg_hash_table[idx_ab5leg]*sph[ld] 
                        B += b5leg_hash_table[idx_ab5leg]*sph[ld] 
                        idx_ab5leg += 1
                                                        
                    A, B = eimph * A, eimph * B
                        
                    for tau1 in range(2):
                        n1 = c_multi_to_single_index(tau1, l1, m1, lmax1, mmax1)
                        for tau2 in range(2):
                            n2 = c_multi_to_single_index(tau2, l2, m2, lmax2, mmax2)
                            if tau1 == tau2:
                                w[n1, n2] = A
                            else:
                                w[n1, n2] = B

    return w
##############################################################################
#  Direct coupling block using  hash table
##############################################################################
@cython.boundscheck(False)
@cython.wraparound(False)
cdef double complex[:,:] c_svwf_translate_2D_from_hash_table_threaded(long blocksize1,
                                                            long blocksize2,
                                                            double complex[:,:] w,
                                                            double complex[:] sph,
                                                            double complex k,
                                                            double dx,
                                                            double dy,
                                                            double dz,
                                                            long lmax1,
                                                            long lmax2,
                                                            long mmax1,
                                                            long mmax2,
                                                            double complex[:] a5leg_hash_table,
                                                            double complex[:] b5leg_hash_table,
                                                            long kind) nogil:      

                                 


    cdef double phi = atan2(dy, dx)
    cdef double complex d = csqrt(pow(dx,2) + pow(dy,2) + pow(dz,2))

    # If the particle is itself then coupling matrix = 0 matrix.
    if d == 0:
        return w
    
    # spherical functions
    cdef double complex kd = k*d
    
    if kind == 0: # 0 = incoming wave, 1 = outgoing wave. 
    #NOTE: an long flag is used instead of a string because Cython 
    # has troublesome conventions about datatyping strings from python v2 to v3
    # See more on this at https://cython.readthedocs.io/en/latest/src/tutorial/strings.html
        for n in range(lmax1 + lmax2 + 1):
            sph[n] = c_spherical_hankel(n, kd) 
    else:
        for n in range(lmax1 + lmax2 + 1):
            sph[n] = c_spherical_bessel(n, kd)  
  

    # calculate coupling matrix
    cdef long m1, m2, l1, l2, ld, n1, n2, tau1, tau2, idx_ab5leg, idx_eim 
    cdef double complex A, B, eimph
                                                        
    # the particle coupling operator is the transpose of the SVWF translation operator
    # therefore, (l1,m1) and (l2,m2) are interchanged:
    idx_ab5leg = 0       
    for m1 in prange(-mmax1, mmax1 + 1):
        for m2 in range(-mmax2, mmax2 + 1):
            eimph = cexp(1j * (m2 - m1) * phi) 
            for l1 in range(max(1, abs(m1)), lmax1 + 1):
                for l2 in range(max(1, abs(m2)), lmax2 + 1):
                    A = 0
                    B = 0
                    for ld in range(max(abs(l1 - l2), abs(m1 - m2)), l1 + l2 + 1):  # if ld<abs(m1-m2) then P=0
                       
                        A += a5leg_hash_table[idx_ab5leg]*sph[ld] 
                        B += b5leg_hash_table[idx_ab5leg]*sph[ld] 
                        idx_ab5leg += 1
                                                        
                    A, B = eimph * A, eimph * B
                        
                    for tau1 in range(2):
                        n1 = c_multi_to_single_index(tau1, l1, m1, lmax1, mmax1)
                        for tau2 in range(2):
                            n2 = c_multi_to_single_index(tau2, l2, m2, lmax2, mmax2)
                            if tau1 == tau2:
                                w[n1, n2] = A
                            else:
                                w[n1, n2] = B

    return w


# Have option to thread build into the translate function for easier reuse.
def svwf_translate_2D_from_hash_table(long blocksize1,long blocksize2,
                            double complex[:,:] w,
                            double complex[:] sph,
                            double complex k,
                            double dx, double dy, double dz,
                            long lmax1, long lmax2, long mmax1, long mmax2,
                            double complex[:] a5leg_array, 
                            double complex[:] b5leg_array,
                            long kind, threaded = False):    

    cdef double complex [:,:] val
    if threaded == True:
        with nogil:
            val = c_svwf_translate_2D_from_hash_table_threaded(blocksize1,blocksize2,
                                                    w,sph,k,dx, dy, dz,
                                                    lmax1, lmax2, mmax1, mmax2,
                                                    a5leg_array,b5leg_array,
                                                    kind)
    else:
        with nogil:
            val = c_svwf_translate_2D_from_hash_table(blocksize1,blocksize2,
                                            w,sph,k,dx, dy, dz,
                                            lmax1, lmax2, mmax1, mmax2,
                                            a5leg_array,b5leg_array,
                                            kind)
            
    return np.asarray(val, dtype = np.complex128)

# Special case of translate. Make wrapper for readability in smuthi.
def direct_coupling_block_2D_from_hash_table(long blocksize1,long blocksize2,
                            double complex[:,:] w,
                            double complex[:] sph,
                            double complex k,
                            double dx, double dy, double dz,
                            long lmax1, long lmax2, long mmax1, long mmax2,
                            double complex[:] a5leg_array, 
                            double complex[:] b5leg_array,
                            threaded = False):
    cdef long kind = 0
    return svwf_translate_2D_from_hash_table(blocksize1,blocksize2,
                                    w,sph,k,dx, dy, dz,
                                    lmax1, lmax2, mmax1, mmax2,
                                    a5leg_array,b5leg_array,
                                    kind, threaded)



'''
##############################################################################
#****************************************************************************#
#****************************************************************************#
#*******************  Class for managing translations  **********************#
#********************* Includes multithread options *************************#
#****************************************************************************#
##############################################################################

This section defines a class that can more easily abstract all the translation 
options defined above. This could be used as a more readable method to implement
the different options. 

Note: This class is based on the idea that the hash options are always faster
compared to the direct calculation. Also, the initial cost in setup and memory 
of the hash is minimal. Therefore, it should always be used as the default 
option. 

Futhermore, the class is a python class so that getter and setter methods do not
need to be defined in order to pickle. 
'''


# class SWE_DirectTranslation():
    
#     def __init__(self, lmax1, mmax1, lmax2, mmax2):
#         self.lmax1 = int(lmax1)
#         self.lmax2 = int(lmax2)
#         self.mmax1 = int(mmax1)
#         self.mmax2 = int(mmax2)
#         self.a5leg_array = None
#         self.b5leg_array = None
#         self.a5_array = None
#         self.b5_array = None  
        
#         return 
        
#     def is_same(self, lmax1, mmax1, lmax2, mmax2):
#         return (self.lmax1 == lmax1 and self.lmax2 == lmax2 and self.mmax1 == mmax1 and self.mmax2 == mmax2)
        
#     def direct_coupling_block(self, k, dx, dy, dz, threaded = False):
#         harmonic_kind = 'outgoing to regular'
#         return self.translate(self, k, dx, dy, dz,
#                   harmonic_kind, threaded)
        
#     def translate(self, k, dx, dy, dz,
#                   harmonic_kind, threaded = False):
        
#         if harmonic_kind == 'outgoing to regular':
#             kind = int(0)
#         elif harmonic_kind == 'outgoing to outgoing':
#             kind = int(1)
#         else:
#             raise Exception(f'''The svwf translation matrix transforms a particle into a basis
#         of either "incoming" or "outgoing" spherical vector harmonics.
#         \n You did not properly specify if the harmonic_kind should be "incoming" or "outgoing"
#         \n Instead you specified harmonic_kind = {harmonic_kind}''') 
        
#         if dz == 0:
#             return self._translate_2d(k, dx, dy, dz, kind, threaded)
#         else: 
#             return self._translate_3d(k, dx, dy, dz, kind, threaded)
        

#     def __translate_2d(self, k, dx, dy, dz, kind, threaded):
        
#         if (self.a5leg_array == None or self.b5leg_array == None):
#             self.a5leg_array, self.b5leg_array = c_hard_ab5_coefficient_and_legendre_hash_table(self.lmax1, 
#                                                                                             self.lmax2, 
#                                                                                             self.mmax1, 
#                                                                                             self.mmax2)
#         return csvwf_translate_2D_from_hash_table(complex(k),float(dx),float(dy),float(dz),
#                          self.lmax1,self.lmax2,self.mmax1,self.mmax2,
#                          self.a5leg_hash_table, self.b5leg_hash_table,
#                          kind, threaded)  
    
#     def __translate_3d(self, k, dx, dy, dz, kind, threaded):
        
#         if (self.a5_array == None or self.b5_array == None):
#             self.a5_array, self.b5_array = chard_ab5_coefficient_hash_table(self.lmax1, 
#                                                                         self.lmax2, 
#                                                                         self.mmax1, 
#                                                                         self.mmax2)
            
#         return csvwf_translate_3D_from_hash_table(complex(k),float(dx),float(dy),float(dz),
#                          self.lmax1,self.lmax2,self.mmax1,self.mmax2,
#                          self.a5_hash_table, self.b5_hash_table,
#                          kind, threaded)  
        























"""Functions for the transformation of plane and spherical vector wave 
functions as well as of plane and spherical wave fex."""

import numpy as np
import smuthi.fields.expansions as fex
import smuthi.fields as flds
import smuthi.utility.math as mathfunc
import smuthi.utility.memoizing as memo
from smuthi.fields.spherical_harmonic_translations import svwf_translate



###############################################################################
#                          transformations                                    #
###############################################################################


def transformation_coefficients_vwf(tau, l, m, pol, kp=None, kz=None, pilm_list=None, taulm_list=None, dagger=False):
    r"""Transformation coefficients B to expand SVWF in PVWF and vice versa:

    .. math::
        B_{\tau l m,j}(x) = -\frac{1}{\mathrm{i}^{l+1}} \frac{1}{\sqrt{2l(l+1)}} (\mathrm{i} \delta_{j1} + \delta_{j2})
        (\delta_{\tau j} \tau_l^{|m|}(x) + (1-\delta_{\tau j} m \pi_l^{|m|}(x))

    For the definition of the :math:`\tau_l^m` and :math:`\pi_l^m` functions, see
    `A. Doicu, T. Wriedt, and Y. A. Eremin: "Light Scattering by Systems of Particles", Springer-Verlag, 2006
    <https://doi.org/10.1007/978-3-540-33697-6>`_
    
    Compare also section 2.3.3 of [Egel 2018 diss].
    
    Args:
        tau (int):          SVWF polarization, 0 for spherical TE, 1 for spherical TM
        l (int):            l=1,... SVWF multipole degree
        m (int):            m=-l,...,l SVWF multipole order
        pol (int):          PVWF polarization, 0 for TE, 1 for TM
        kp (numpy array):         PVWF in-plane wavenumbers
        kz (numpy array):         complex numpy-array: PVWF out-of-plane wavenumbers
        pilm_list (list):   2D list numpy-arrays: alternatively to kp and kz, pilm and taulm as generated with
                            legendre_normalized can directly be handed
        taulm_list (list):  2D list numpy-arrays: alternatively to kp and kz, pilm and taulm as generated with
                            legendre_normalized can directly be handed
        dagger (bool):      switch on when expanding PVWF in SVWF and off when expanding SVWF in PVWF

    Returns:
        Transformation coefficient as array (size like kp).
    """
    if pilm_list is None:
        k = np.sqrt(kp**2 + kz**2)
        ct = kz / k
        st = kp / k
        plm_list, pilm_list, taulm_list = mathfunc.legendre_normalized(ct, st, l)

    if tau == pol:
        sphfun = taulm_list[l, abs(m)]
    else:
        sphfun = m * pilm_list[l, abs(m)]

    if dagger:
        if pol == 0:
            prefac = -1 / (-1j) ** (l + 1) / np.sqrt(2 * l * (l + 1)) * (-1j)
        elif pol == 1:
            prefac = -1 / (-1j) ** (l + 1) / np.sqrt(2 * l * (l + 1)) * 1
        else:
            raise ValueError('pol must be 0 (TE) or 1 (TM)')
    else:
        if pol == 0:
            prefac = -1 / (1j) ** (l + 1) / np.sqrt(2 * l * (l + 1)) * (1j)
        elif pol ==1:
            prefac = -1 / (1j) ** (l + 1) / np.sqrt(2 * l * (l + 1)) * 1
        else:
            raise ValueError('pol must be 0 (TE) or 1 (TM)')

    B = prefac * sphfun
    return B


def pwe_to_swe_conversion(pwe, l_max, m_max, reference_point):
    """Convert plane wave expansion object to a spherical wave expansion object.

    Args:
        pwe (PlaneWaveExpansion):   Plane wave expansion to be converted
        l_max (int):                Maximal multipole degree of spherical wave expansion
        m_max (int):                Maximal multipole order of spherical wave expansion
        reference_point (list):     Coordinates of reference point in the format [x, y, z]

    Returns:
        SphericalWaveExpansion object.
    """

    if reference_point[2] < pwe.lower_z or reference_point[2] > pwe.upper_z:
        raise ValueError('reference point not inside domain of pwe validity')

    swe = fex.SphericalWaveExpansion(k=pwe.k, l_max=l_max, m_max=m_max, kind='regular', 
                                 reference_point=reference_point, lower_z=pwe.lower_z, 
                                 upper_z=pwe.upper_z)
    kpgrid = pwe.k_parallel_grid()
    agrid = pwe.azimuthal_angle_grid()
    kx = kpgrid * np.cos(agrid)
    ky = kpgrid * np.sin(agrid)
    kz = pwe.k_z_grid()
    kzvec = pwe.k_z()

    kvec = np.array([kx, ky, kz])
    rswe_mn_rpwe = np.array(reference_point) - np.array(pwe.reference_point)

    # phase factor for the translation of the reference point from rvec_iS to rvec_S
    ejkriSS = np.exp(1j * (kvec[0] * rswe_mn_rpwe[0] + kvec[1] * rswe_mn_rpwe[1] + kvec[2] * rswe_mn_rpwe[2]))

    # phase factor times pwe coefficients
    gejkriSS = pwe.coefficients * ejkriSS[None, :, :]  # indices: pol, jk, ja
    
    ct = kzvec / pwe.k
    st = pwe.k_parallel / pwe.k
    plm_list, pilm_list, taulm_list = mathfunc.legendre_normalized(ct, st, l_max)

    for m in range(-m_max, m_max + 1):
        emjma_geijkriSS = np.exp(-1j * m * pwe.azimuthal_angles)[None, None, :] * gejkriSS
        for l in range(max(1, abs(m)), l_max + 1):
            for tau in range(2):
                ak_integrand = np.zeros(kpgrid.shape, dtype=complex)
                for pol in range(2):
                    Bdag = transformation_coefficients_vwf(tau, l, m, pol=pol, pilm_list=pilm_list,
                                                               taulm_list=taulm_list, kz=kzvec, dagger=True)
                    ak_integrand += Bdag[:, None] * emjma_geijkriSS[pol, :, :]
                if len(pwe.k_parallel) > 1:
                    an = np.trapz(np.trapz(ak_integrand, pwe.azimuthal_angles) * pwe.k_parallel, pwe.k_parallel) * 4
                else:
                    an = ak_integrand * 4
                swe.coefficients[flds.multi_to_single_index(tau, l, m, swe.l_max, swe.m_max)] = np.squeeze(an)
    return swe


def swe_to_pwe_conversion(swe, k_parallel, azimuthal_angles, layer_system=None, layer_number=None,
                          layer_system_mediated=False,
                          only_l=None, only_m=None, only_pol=None, only_tau=None):
    """Convert SphericalWaveExpansion object to a PlaneWaveExpansion object.

    Args:
        swe (SphericalWaveExpansion):             Spherical wave expansion to be converted
        k_parallel (numpy array or str):          In-plane wavenumbers for the pwe object.
        azimuthal_angles (numpy array or str):    Azimuthal angles for the pwe object
        layer_system (smuthi.layers.LayerSystem): Stratified medium in which the origin of the SWE is located
        layer_number (int):                       Layer number in which the PWE should be valid.
        layer_system_mediated (bool):             If True, the PWE refers to the layer system response of the SWE, 
                                                  otherwise it is the direct transform.
        only_pol (int):  if set to 0 or 1, only this plane wave polarization (0=TE, 1=TM) is considered
        only_tau (int):  if set to 0 or 1, only this spherical vector wave polarization (0 — magnetic, 1 — electric) is
                         considered
        only_l (int):    if set to positive number, only this multipole degree is considered
        only_m (int):    if set to non-negative number, only this multipole order is considered

    Returns:
        Tuple of two PlaneWaveExpansion objects, first upgoing, second downgoing.
    """
    # todo: manage diverging swe
    k_parallel = np.array(k_parallel, ndmin=1)
    
    i_swe = layer_system.layer_number(swe.reference_point[2])
    if layer_number is None and not layer_system_mediated:
        layer_number = i_swe
    reference_point = [0, 0, layer_system.reference_z(i_swe)]
    lower_z_up = swe.reference_point[2]
    upper_z_up = layer_system.upper_zlimit(layer_number)
    pwe_up = fex.PlaneWaveExpansion(k=swe.k, k_parallel=k_parallel, azimuthal_angles=azimuthal_angles, kind='upgoing',
                                    reference_point=reference_point, lower_z=lower_z_up, upper_z=upper_z_up)
    lower_z_down = layer_system.lower_zlimit(layer_number)
    upper_z_down = swe.reference_point[2]
    pwe_down = fex.PlaneWaveExpansion(k=swe.k, k_parallel=k_parallel, azimuthal_angles=azimuthal_angles,
                                      kind='downgoing', reference_point=reference_point, lower_z=lower_z_down,
                                      upper_z=upper_z_down)

    agrid = pwe_up.azimuthal_angle_grid()
    kpgrid = pwe_up.k_parallel_grid()
    kx = kpgrid * np.cos(agrid)
    ky = kpgrid * np.sin(agrid)
    kz_up = pwe_up.k_z_grid()
    kz_down = pwe_down.k_z_grid()

    kzvec = pwe_up.k_z()

    kvec_up = np.array([kx, ky, kz_up])
    kvec_down = np.array([kx, ky, kz_down])
    rpwe_mn_rswe = np.array(reference_point) - np.array(swe.reference_point)

    # phase factor for the translation of the reference point from rvec_S to rvec_iS
    ejkrSiS_up = np.exp(1j * np.tensordot(kvec_up, rpwe_mn_rswe, axes=([0], [0])))
    ejkrSiS_down = np.exp(1j * np.tensordot(kvec_down, rpwe_mn_rswe, axes=([0], [0])))
    
    ct_up = pwe_up.k_z() / swe.k
    st_up = pwe_up.k_parallel / swe.k
    plm_list_up, pilm_list_up, taulm_list_up = mathfunc.legendre_normalized(ct_up, st_up, swe.l_max)

    ct_down = pwe_down.k_z() / swe.k
    st_down = pwe_down.k_parallel / swe.k
    plm_list_down, pilm_list_down, taulm_list_down = mathfunc.legendre_normalized(ct_down, st_down, swe.l_max)
    
    for m in range(-swe.m_max, swe.m_max + 1):
        if only_m is not None and m != only_m:
            continue
        eima = np.exp(1j * m * pwe_up.azimuthal_angles)  # indices: alpha_idx
        for pol in range(2):
            if only_pol is not None and pol != only_pol:
                continue
            dbB_up = np.zeros(len(k_parallel), dtype=complex)
            dbB_down = np.zeros(len(k_parallel), dtype=complex)
            for l in range(max(1, abs(m)), swe.l_max + 1):
                if only_l is not None and l != only_l:
                    continue
                for tau in range(2):
                    if only_tau is not None and tau != only_tau:
                        continue
                    dbB_up += swe.coefficients_tlm(tau, l, m) * transformation_coefficients_vwf(
                        tau, l, m, pol, pilm_list=pilm_list_up, taulm_list=taulm_list_up)
                    dbB_down += swe.coefficients_tlm(tau, l, m) * transformation_coefficients_vwf(
                        tau, l, m, pol, pilm_list=pilm_list_down, taulm_list=taulm_list_down)
            pwe_up.coefficients[pol, :, :] += dbB_up[:, None] * eima[None, :]
            pwe_down.coefficients[pol, :, :] += dbB_down[:, None] * eima[None, :]

    pwe_up.coefficients = pwe_up.coefficients / (2 * np.pi * kzvec[None, :, None] * swe.k) * ejkrSiS_up[None, :, :]
    pwe_down.coefficients = (pwe_down.coefficients / (2 * np.pi * kzvec[None, :, None] * swe.k)
                             * ejkrSiS_down[None, :, :])

    if layer_system_mediated:
        pwe_up, pwe_down = layer_system.response((pwe_up, pwe_down), i_swe, layer_number)

    return pwe_up, pwe_down


###############################################################################
#                          translations                                       #
###############################################################################


def translation_block(vacuum_wavelength, receiving_particle, emitting_particle, layer_system, kind):
    r"""Direct particle translation matrix :math:`W` for two particles that do not have intersecting circumscribing spheres.
       This routine is explicit.
       
       To reduce computation time, this routine relies on two internal accelerations. 
       First, in most cases the number of unique maximum multipole indicies,
       :math:`(\tau, l_{max}, m_{max})`, is much less than the number of unique particles. 
       Therefore, all calculations that depend only on multipole indicies are stored in an 
       intermediate hash table. Second, Cython acceleration is used by default to leverage 
       fast looping. If the Cython files are not supported, this routine will 
       fall back on equivalent Python looping.
       
       Cython acceleration can be between 10-1,000x faster compared to the Python 
       equivalent. Speed variability depends on the number of unique multipoles indicies,
       the size of the largest multipole order, and if particles share the same z coordinate.
       

    Args:
        vacuum_wavelength (float):                          Vacuum wavelength :math:`\lambda` (length unit)
        receiving_particle (smuthi.particles.Particle):     Particle that receives the scattered field
        emitting_particle (smuthi.particles.Particle):      Particle that emits the scattered field
        layer_system (smuthi.layers.LayerSystem):           Stratified medium in which the coupling takes place

    Returns:
        Direct coupling matrix block as numpy array.
    """
    
    # Note: Direct coupling is transpose of translation. 
    # The translation convention is emit = 1 -> recieve = 2
    # The transpose of this is emit = 2 -> recieve = 1
    lmax1 = int(emitting_particle.l_max) 
    mmax1 = int(emitting_particle.m_max)
    lmax2 = int(receiving_particle.l_max)
    mmax2 = int(receiving_particle.m_max)    

    
    # Check if particles are in the same layer.
    rS1 = emitting_particle.position
    rS2 = receiving_particle.position
    iS1 = layer_system.layer_number(rS1[2])
    iS2 = layer_system.layer_number(rS2[2])
    
    if iS1 == iS2 or layer_system.is_degenerate():
        if not emitting_particle == receiving_particle:
            # Initialize variables from abstract classes to be passed as simple 
            # data types
            omega = flds.angular_frequency(vacuum_wavelength)
            k = complex(omega * layer_system.refractive_indices[iS1])
            d = [rS1[i]-rS2[i] for i in range(3)]
            return svwf_translate(k, d, lmax1, mmax1, lmax2, mmax2, kind, threaded = False)
        else:
            blocksize1 = int(flds.blocksize(lmax1, mmax1)) # Rows of translation matrix
            blocksize2 = int(flds.blocksize(lmax2, mmax2)) # Cols of translation matrix
            # Note Translation returns identity such that you recieve the same coefficients on multiplication (translate to self returns self..)
            return np.eye((blocksize1, blocksize2), dtype = np.complex128, order = 'C') 
    else:
        raise Exception('''
                        You are attempting to directly translate a particle through a layer interface. 
                        Direct translation (2D or 3D) is only sensible for spaces in the same layer.
                        You need to first deal with the effect of the layer interface!
                        ''')


def translation_coefficients_svwf(tau1, l1, m1, tau2, l2, m2, k, d, sph_hankel=None, legendre=None, exp_immphi=None):
    r"""Coefficients of the translation operator for the expansion of an outgoing spherical wave in terms of
    regular spherical waves with respect to a different origin:

    .. math::
        \mathbf{\Psi}_{\tau l m}^{(3)}(\mathbf{r} + \mathbf{d} = \sum_{\tau'} \sum_{l'} \sum_{m'}
        A_{\tau l m, \tau' l' m'} (\mathbf{d}) \mathbf{\Psi}_{\tau' l' m'}^{(1)}(\mathbf{r})

    for :math:`|\mathbf{r}|<|\mathbf{d}|`.
    
    See also section 2.3.3 and appendix B of [Egel 2018 diss].
    
    Args:
        tau1 (int):             tau1=0,1: Original wave's spherical polarization
        l1 (int):               l=1,...: Original wave's SVWF multipole degree
        m1 (int):               m=-l,...,l: Original wave's SVWF multipole order
        tau2 (int):             tau2=0,1: Partial wave's spherical polarization
        l2 (int):               l=1,...: Partial wave's SVWF multipole degree
        m2 (int):               m=-l,...,l: Partial wave's SVWF multipole order
        k (float or complex):   wavenumber (inverse length unit)
        d (list):               translation vectors in format [dx, dy, dz] (length unit)
                                dx, dy, dz can be scalars or ndarrays
        sph_hankel (list):      Optional. sph_hankel[i] contains the spherical hankel funciton of degree i, evaluated at
                                k*d where d is the norm of the distance vector(s)
        legendre (list):        Optional. legendre[l][m] contains the legendre function of order l and degree m,
                                evaluated at cos(theta) where theta is the polar angle(s) of the distance vector(s)

    Returns:
        translation coefficient A (complex)

    """
    # spherical coordinates of d:
    dd = np.sqrt(d[0] ** 2 + d[1] ** 2 + d[2] ** 2)

    if exp_immphi is None:
        phid = np.arctan2(d[1], d[0])
        eimph = np.exp(1j * (m1 - m2) * phid)
    else:
        eimph = exp_immphi[m1][m2]

    if sph_hankel is None:
        sph_hankel = [mathfunc.spherical_hankel(n, k * dd) for n in range(l1 + l2 + 1)]

    if legendre is None:
        costthetd = d[2] / dd
        sinthetd = np.sqrt(d[0] ** 2 + d[1] ** 2) / dd
        legendre, _, _ = mathfunc.legendre_normalized(costthetd, sinthetd, l1 + l2)

    A = complex(0)
    for ld in range(abs(l1 - l2), l1 + l2 + 1):
        a5, b5 = ab5_coefficients(l1, m1, l2, m2, ld)
        if tau1 == tau2:
            A += a5 * sph_hankel[ld] * legendre[ld, abs(m1 - m2)]
        else:
            A += b5 * sph_hankel[ld] * legendre[ld, abs(m1 - m2)]
    A = eimph * A
    return A


def translation_coefficients_svwf_out_to_out(tau1, l1, m1, tau2, l2, m2, k, d, sph_bessel=None, legendre=None,
                                             exp_immphi=None):
    r"""Coefficients of the translation operator for the expansion of an outgoing spherical wave in terms of
    outgoing spherical waves with respect to a different origin:

    .. math::
        \mathbf{\Psi}_{\tau l m}^{(3)}(\mathbf{r} + \mathbf{d} = \sum_{\tau'} \sum_{l'} \sum_{m'}
        A_{\tau l m, \tau' l' m'} (\mathbf{d}) \mathbf{\Psi}_{\tau' l' m'}^{(3)}(\mathbf{r})

    for :math:`|\mathbf{r}|>|\mathbf{d}|`.

    Args:
        tau1 (int):             tau1=0,1: Original wave's spherical polarization
        l1 (int):               l=1,...: Original wave's SVWF multipole degree
        m1 (int):               m=-l,...,l: Original wave's SVWF multipole order
        tau2 (int):             tau2=0,1: Partial wave's spherical polarization
        l2 (int):               l=1,...: Partial wave's SVWF multipole degree
        m2 (int):               m=-l,...,l: Partial wave's SVWF multipole order
        k (float or complex):   wavenumber (inverse length unit)
        d (list):               translation vectors in format [dx, dy, dz] (length unit)
                                dx, dy, dz can be scalars or ndarrays
        sph_bessel (list):      Optional. sph_bessel[i] contains the spherical Bessel funciton of degree i, evaluated at
                                k*d where d is the norm of the distance vector(s)
        legendre (list):        Optional. legendre[l][m] contains the legendre function of order l and degree m,
                                evaluated at cos(theta) where theta is the polar angle(s) of the distance vector(s)

    Returns:
        translation coefficient A (complex)
    """
    # spherical coordinates of d:
    dd = np.sqrt(d[0] ** 2 + d[1] ** 2 + d[2] ** 2)

    if exp_immphi is None:
        phid = np.arctan2(d[1], d[0])
        eimph = np.exp(1j * (m1 - m2) * phid)
    else:
        eimph = exp_immphi[m1][m2]

    if sph_bessel is None:
        sph_bessel = [mathfunc.spherical_bessel(n, k * dd) for n in range(l1 + l2 + 1)]

    if legendre is None:
        costthetd = d[2] / dd
        sinthetd = np.sqrt(d[0] ** 2 + d[1] ** 2) / dd
        legendre, _, _ = mathfunc.legendre_normalized(costthetd, sinthetd, l1 + l2)

    A = complex(0)
    for ld in range(abs(l1 - l2), l1 + l2 + 1):
        a5, b5 = ab5_coefficients(l1, m1, l2, m2, ld)
        if tau1==tau2:
            A += a5 * sph_bessel[ld] * legendre[ld, abs(m1 - m2)]
        else:
            A += b5 * sph_bessel[ld] * legendre[ld, abs(m1 - m2)]
    A = eimph * A
    return A


@memo.Memoize
# @jit(complex128[:](int32, int32, int32, int32, int32),
#      nopython=True, cache=True)
def ab5_coefficients(l1, m1, l2, m2, p):
    """a5 and b5 are the coefficients used in the evaluation of the SVWF translation
    operator.
    
    See section 2.3.3 and appendix B of [Egel 2018 diss].

    Args:
        l1 (int):           l=1,...: Original wave's SVWF multipole degree
        m1 (int):           m=-l,...,l: Original wave's SVWF multipole order
        l2 (int):           l=1,...: Partial wave's SVWF multipole degree
        m2 (int):           m=-l,...,l: Partial wave's SVWF multipole order
        p (int):            p parameter

    Returns:
        A tuple (a5, b5) where a5 and b5 are complex.
    """
    jfac = 1.0j ** (abs(m1 - m2) - abs(m1) - abs(m2) + l2 - l1 + p) * (-1) ** (m1 - m2)
    fac1 = np.sqrt((2 * l1 + 1) * (2 * l2 + 1) / (2 * l1 * (l1 + 1) * l2 * (l2 + 1)))
    fac2a = (l1 * (l1 + 1) + l2 * (l2 + 1) - p * (p + 1)) * np.sqrt(2 * p + 1)
    fac2b = np.sqrt((l1 + l2 + 1 + p) * (l1 + l2 + 1 - p) * (p + l1 - l2) * (p - l1 + l2) * (2 * p + 1))
    # Note that arguments are in two_j = 2*j.
    wig1 = mathfunc.nb_wig3jj(2*l1, 2*l2, 2*p, 2*m1, -m2*2, -(m1 - m2)*2)
    wig2a = mathfunc.nb_wig3jj(2*l1, 2*l2, 2*p, 0, 0, 0)
    wig2b = mathfunc.nb_wig3jj(2*l1, 2*l2, 2*(p - 1), 0, 0, 0)

    a = jfac * fac1 * fac2a * wig1 * wig2a
    b = jfac * fac1 * fac2b * wig1 * wig2b
    return np.array([a, b])


###############################################################################
#                          rotations                                          #
###############################################################################

def block_rotation_matrix_D_svwf(l_max, m_max, alpha, beta, gamma, wdsympy=False):
    """Rotation matrix for the rotation of SVWFs between the labratory 
    coordinate system (L) and a rotated coordinate system (R)
    
    Args:
        l_max (int):      Maximal multipole degree
        m_max (int):      Maximal multipole order
        alpha (float):    First Euler angle, rotation around z-axis, in rad
        beta (float):     Second Euler angle, rotation around y'-axis in rad
        gamma (float):    Third Euler angle, rotation around z''-axis in rad
        wdsympy (bool):   If True, Wigner-d-functions come from the sympy toolbox
        
    Returns:
        rotation matrix of dimension [blocksize, blocksize]
    """
    
    b_size = flds.blocksize(l_max, m_max)
    rotation_matrix = np.zeros([b_size, b_size], dtype=complex)
    
    for l in range(l_max + 1):
        mstop = min(l, m_max)
        for m1 in range(-mstop, mstop + 1):
            for m2 in range(-mstop, mstop + 1):
                rotation_matrix_coefficient = mathfunc.wigner_D(l, m1, m2, alpha, beta, gamma, wdsympy)
                for tau in range(2):
                    n1 = flds.multi_to_single_index(tau, l, m1, l_max, m_max)
                    n2 = flds.multi_to_single_index(tau, l, m2, l_max, m_max)
                    rotation_matrix[n1, n2] = rotation_matrix_coefficient

    return rotation_matrix



"""This module contains CUDA source code for the evaluation of the electric 
field from a VWF expansion."""

# This cuda kernel is used for the evaluation of the electric field of plane wave expansions.
pwe_electric_field_evaluation_code = """
    #define LEN_X %i
    #define LEN_K %i
    #define LEN_A %i
    #define RE_ONE_OVER_K %.10f
    #define IM_ONE_OVER_K %.10f
    
    __device__ void alpha_integral(const int i_k, const float x, const float y, const float z, 
                                   const float re_kp, const float im_kp, const float re_kz, const float im_kz, 
                                   const float *alpha_array, 
                                   const float *re_g_te_array, const float *im_g_te_array,
                                   const float *re_g_tm_array, const float *im_g_tm_array,
                                   float *re_integral_x, float *im_integral_x,
                                   float *re_integral_y, float *im_integral_y,
                                   float *re_integral_z, float *im_integral_z)
    {
        float a = 0.0;
        
        float re_x_result = 0.0;
        float im_x_result = 0.0;
        float re_y_result = 0.0;
        float im_y_result = 0.0;
        float re_z_result = 0.0;
        float im_z_result = 0.0;
        
        float re_e_x_alpha_integrand = 0.0;
        float im_e_x_alpha_integrand = 0.0;
        float re_e_y_alpha_integrand = 0.0;
        float im_e_y_alpha_integrand = 0.0;
        float re_e_z_alpha_integrand = 0.0;
        float im_e_z_alpha_integrand = 0.0;

        for (int i_a=0; i_a<LEN_A; i_a++)
        {
            float a_old = a;
            a = alpha_array[i_a];
            float d_a = a - a_old;
            float sin_a = sinf(a);
            float cos_a = cosf(a);
            
            float re_ikr = -im_kp * (cos_a * x + sin_a * y) - im_kz * z;
            float im_ikr = re_kp * (cos_a * x + sin_a * y) + re_kz * z;
            
            float ereikr = expf(re_ikr);
            float re_eikr =  ereikr * cosf(im_ikr);
            float im_eikr =  ereikr * sinf(im_ikr);
            
            // integrand
            float re_e_x_alpha_integrand_old = re_e_x_alpha_integrand;
            float im_e_x_alpha_integrand_old = im_e_x_alpha_integrand;
            float re_e_y_alpha_integrand_old = re_e_y_alpha_integrand;
            float im_e_y_alpha_integrand_old = im_e_y_alpha_integrand;
            float re_e_z_alpha_integrand_old = re_e_z_alpha_integrand;
            float im_e_z_alpha_integrand_old = im_e_z_alpha_integrand;
            
            int i_ka = i_k * LEN_A + i_a;
            
            // pol=TE
            float re_g = re_g_te_array[i_ka];
            float im_g = im_g_te_array[i_ka];

            float re_geikr = re_g * re_eikr - im_g * im_eikr;
            float im_geikr = re_g * im_eikr + im_g * re_eikr;
            
            re_e_x_alpha_integrand = -sin_a * re_geikr;
            im_e_x_alpha_integrand = -sin_a * im_geikr;
            re_e_y_alpha_integrand = cos_a * re_geikr;
            im_e_y_alpha_integrand = cos_a * im_geikr;

            // pol=TM
            re_g = re_g_tm_array[i_ka];
            im_g = im_g_tm_array[i_ka];

            re_geikr = re_g * re_eikr - im_g * im_eikr;
            im_geikr = re_g * im_eikr + im_g * re_eikr;
            
            float re_kzgeikr = re_kz * re_geikr - im_kz * im_geikr;
            float im_kzgeikr = re_kz * im_geikr + im_kz * re_geikr;
            
            float re_kzkgeikr = re_kzgeikr * RE_ONE_OVER_K - im_kzgeikr * IM_ONE_OVER_K;
            float im_kzkgeikr = re_kzgeikr * IM_ONE_OVER_K + im_kzgeikr * RE_ONE_OVER_K;
            
            re_e_x_alpha_integrand += cos_a * re_kzkgeikr;
            im_e_x_alpha_integrand += cos_a * im_kzkgeikr;
            re_e_y_alpha_integrand += sin_a * re_kzkgeikr;
            im_e_y_alpha_integrand += sin_a * im_kzkgeikr;

            float re_kpgeikr = re_kp * re_geikr - im_kp * im_geikr;
            float im_kpgeikr = re_kp * im_geikr + im_kp * re_geikr;
            re_e_z_alpha_integrand = -(re_kpgeikr * RE_ONE_OVER_K - im_kpgeikr * IM_ONE_OVER_K);
            im_e_z_alpha_integrand = -(re_kpgeikr * IM_ONE_OVER_K + im_kpgeikr * RE_ONE_OVER_K);
            
            if (i_a>0)
            {
                re_x_result += 0.5 * d_a * (re_e_x_alpha_integrand + re_e_x_alpha_integrand_old);
                im_x_result += 0.5 * d_a * (im_e_x_alpha_integrand + im_e_x_alpha_integrand_old);
                re_y_result += 0.5 * d_a * (re_e_y_alpha_integrand + re_e_y_alpha_integrand_old);
                im_y_result += 0.5 * d_a * (im_e_y_alpha_integrand + im_e_y_alpha_integrand_old);
                re_z_result += 0.5 * d_a * (re_e_z_alpha_integrand + re_e_z_alpha_integrand_old);
                im_z_result += 0.5 * d_a * (im_e_z_alpha_integrand + im_e_z_alpha_integrand_old);
            }
        }
        re_integral_x[0] = re_x_result;
        im_integral_x[0] = im_x_result;
        re_integral_y[0] = re_y_result;
        im_integral_y[0] = im_y_result;
        re_integral_z[0] = re_z_result;
        im_integral_z[0] = im_z_result;
    }
    
    
    __global__ void electric_field(const float *re_kp_array, const float *im_kp_array, 
                                   const float *re_kz_array, const float *im_kz_array,
                                   const float *alpha_array,  
                                   const float *x_array, const float *y_array, const float *z_array,
                                   const float *re_g_te_array, const float *im_g_te_array,
                                   const float *re_g_tm_array, const float *im_g_tm_array,
                                   float *re_e_x, float *im_e_x, float *re_e_y, float *im_e_y, 
                                   float *re_e_z, float *im_e_z)
    {
        unsigned int i = blockIdx.x * blockDim.x + threadIdx.x;
        if (i >= LEN_X) return;
        
        float x = x_array[i];
        float y = y_array[i];
        float z = z_array[i];
        
        float re_kp = 0.0;
        float im_kp = 0.0;

        float re_x_result = 0.0;
        float im_x_result = 0.0;
        float re_y_result = 0.0;
        float im_y_result = 0.0;
        float re_z_result = 0.0;
        float im_z_result = 0.0;
        
        float re_x_kappa_integrand = 0.0;
        float im_x_kappa_integrand = 0.0;
        float re_y_kappa_integrand = 0.0;
        float im_y_kappa_integrand = 0.0;
        float re_z_kappa_integrand = 0.0;
        float im_z_kappa_integrand = 0.0;
        
        for (int i_k=0; i_k<LEN_K; i_k++)
        {
            float re_kp_old = re_kp;
            float im_kp_old = im_kp;
            
            re_kp = re_kp_array[i_k];
            im_kp = im_kp_array[i_k];
            
            float re_d_kp = re_kp - re_kp_old;
            float im_d_kp = im_kp - im_kp_old;

            float re_kz = re_kz_array[i_k];
            float im_kz = im_kz_array[i_k];
            
            float re_x_kappa_integrand_old = re_x_kappa_integrand;
            float im_x_kappa_integrand_old = im_x_kappa_integrand;
            float re_y_kappa_integrand_old = re_y_kappa_integrand;
            float im_y_kappa_integrand_old = im_y_kappa_integrand;
            float re_z_kappa_integrand_old = re_z_kappa_integrand;
            float im_z_kappa_integrand_old = im_z_kappa_integrand;
            
            float re_x_alpha_itgrl = 0.0;
            float im_x_alpha_itgrl = 0.0;
            float re_y_alpha_itgrl = 0.0;
            float im_y_alpha_itgrl = 0.0;
            float re_z_alpha_itgrl = 0.0;
            float im_z_alpha_itgrl = 0.0;
            
            alpha_integral(i_k, x, y, z, re_kp, im_kp, re_kz, im_kz, alpha_array, 
                           re_g_te_array, im_g_te_array, re_g_tm_array, im_g_tm_array,
                           &re_x_alpha_itgrl, &im_x_alpha_itgrl, &re_y_alpha_itgrl, &im_y_alpha_itgrl, 
                           &re_z_alpha_itgrl, &im_z_alpha_itgrl);
                           
            re_x_kappa_integrand = re_x_alpha_itgrl * re_kp - im_x_alpha_itgrl * im_kp;
            im_x_kappa_integrand = re_x_alpha_itgrl * im_kp + im_x_alpha_itgrl * re_kp;
            re_y_kappa_integrand = re_y_alpha_itgrl * re_kp - im_y_alpha_itgrl * im_kp;
            im_y_kappa_integrand = re_y_alpha_itgrl * im_kp + im_y_alpha_itgrl * re_kp;
            re_z_kappa_integrand = re_z_alpha_itgrl * re_kp - im_z_alpha_itgrl * im_kp;
            im_z_kappa_integrand = re_z_alpha_itgrl * im_kp + im_z_alpha_itgrl * re_kp;
            
            if (i_k>0)
            {
                re_x_result += 0.5 * (re_d_kp * (re_x_kappa_integrand + re_x_kappa_integrand_old) 
                                      - im_d_kp * (im_x_kappa_integrand + im_x_kappa_integrand_old));
                im_x_result += 0.5 * (re_d_kp * (im_x_kappa_integrand + im_x_kappa_integrand_old) 
                                      + im_d_kp * (re_x_kappa_integrand + re_x_kappa_integrand_old));
                re_y_result += 0.5 * (re_d_kp * (re_y_kappa_integrand + re_y_kappa_integrand_old) 
                                      - im_d_kp * (im_y_kappa_integrand + im_y_kappa_integrand_old));
                im_y_result += 0.5 * (re_d_kp * (im_y_kappa_integrand + im_y_kappa_integrand_old) 
                                      + im_d_kp * (re_y_kappa_integrand + re_y_kappa_integrand_old));
                re_z_result += 0.5 * (re_d_kp * (re_z_kappa_integrand + re_z_kappa_integrand_old) 
                                      - im_d_kp * (im_z_kappa_integrand + im_z_kappa_integrand_old));
                im_z_result += 0.5 * (re_d_kp * (im_z_kappa_integrand + im_z_kappa_integrand_old) 
                                      + im_d_kp * (re_z_kappa_integrand + re_z_kappa_integrand_old));
            }
        }
        re_e_x[i] = re_x_result;
        im_e_x[i] = im_x_result;
        re_e_y[i] = re_y_result;
        im_e_y[i] = im_y_result;
        re_e_z[i] = re_z_result;
        im_e_z[i] = im_z_result;
    }
"""

# This cuda kernel is used for the evaluation of the magnetic field of plane wave expansions.
pwe_magnetic_field_evaluation_code = """
    #define LEN_X %i
    #define LEN_K %i
    #define LEN_A %i
    #define RE_K %.10f
    #define IM_K %.10f
    
    __device__ void alpha_integral(const int i_k, const float x, const float y, const float z, 
                                   const float re_kp, const float im_kp, const float re_kz, const float im_kz, 
                                   const float *alpha_array, 
                                   const float *re_g_te_array, const float *im_g_te_array,
                                   const float *re_g_tm_array, const float *im_g_tm_array,
                                   float *re_integral_x, float *im_integral_x,
                                   float *re_integral_y, float *im_integral_y,
                                   float *re_integral_z, float *im_integral_z)
    {
        float a = 0.0;
        
        float re_x_result = 0.0;
        float im_x_result = 0.0;
        float re_y_result = 0.0;
        float im_y_result = 0.0;
        float re_z_result = 0.0;
        float im_z_result = 0.0;
        
        float re_h_x_alpha_integrand = 0.0;
        float im_h_x_alpha_integrand = 0.0;
        float re_h_y_alpha_integrand = 0.0;
        float im_h_y_alpha_integrand = 0.0;
        float re_h_z_alpha_integrand = 0.0;
        float im_h_z_alpha_integrand = 0.0;

        for (int i_a=0; i_a<LEN_A; i_a++)
        {
            float a_old = a;
            a = alpha_array[i_a];
            float d_a = a - a_old;
            float sin_a = sinf(a);
            float cos_a = cosf(a);
            
            float re_ikr = -im_kp * (cos_a * x + sin_a * y) - im_kz * z;
            float im_ikr = re_kp * (cos_a * x + sin_a * y) + re_kz * z;
            
            float ereikr = expf(re_ikr);
            float re_eikr =  ereikr * cosf(im_ikr);
            float im_eikr =  ereikr * sinf(im_ikr);
            
            // integrand
            float re_h_x_alpha_integrand_old = re_h_x_alpha_integrand;
            float im_h_x_alpha_integrand_old = im_h_x_alpha_integrand;
            float re_h_y_alpha_integrand_old = re_h_y_alpha_integrand;
            float im_h_y_alpha_integrand_old = im_h_y_alpha_integrand;
            float re_h_z_alpha_integrand_old = re_h_z_alpha_integrand;
            float im_h_z_alpha_integrand_old = im_h_z_alpha_integrand;
            
            int i_ka = i_k * LEN_A + i_a;
            
            // pol=TE
            float re_g = re_g_te_array[i_ka];
            float im_g = im_g_te_array[i_ka];

            float re_geikr = re_g * re_eikr - im_g * im_eikr;
            float im_geikr = re_g * im_eikr + im_g * re_eikr;
            
            float re_kzgeikr = re_kz * re_geikr - im_kz * im_geikr;
            float im_kzgeikr = re_kz * im_geikr + im_kz * re_geikr;
            
            re_h_x_alpha_integrand = -cos_a * re_kzgeikr;
            im_h_x_alpha_integrand = -cos_a * im_kzgeikr;
            re_h_y_alpha_integrand = -sin_a * re_kzgeikr;
            im_h_y_alpha_integrand = -sin_a * im_kzgeikr;
            
            re_h_z_alpha_integrand = re_kp * re_geikr - im_kp * im_geikr;
            im_h_z_alpha_integrand = re_kp * im_geikr + im_kp * re_geikr;
            
            // pol=TM
            re_g = re_g_tm_array[i_ka];
            im_g = im_g_tm_array[i_ka];

            re_geikr = re_g * re_eikr - im_g * im_eikr;
            im_geikr = re_g * im_eikr + im_g * re_eikr;
            
            float re_kgeikr = RE_K * re_geikr - IM_K * im_geikr;
            float im_kgeikr = RE_K * im_geikr + IM_K * re_geikr;
            
            re_h_x_alpha_integrand += -sin_a * re_kgeikr;
            im_h_x_alpha_integrand += -sin_a * im_kgeikr;
            re_h_y_alpha_integrand += cos_a * re_kgeikr;
            im_h_y_alpha_integrand += cos_a * im_kgeikr;
                       
            
            if (i_a>0)
            {
                re_x_result += 0.5 * d_a * (re_h_x_alpha_integrand + re_h_x_alpha_integrand_old);
                im_x_result += 0.5 * d_a * (im_h_x_alpha_integrand + im_h_x_alpha_integrand_old);
                re_y_result += 0.5 * d_a * (re_h_y_alpha_integrand + re_h_y_alpha_integrand_old);
                im_y_result += 0.5 * d_a * (im_h_y_alpha_integrand + im_h_y_alpha_integrand_old);
                re_z_result += 0.5 * d_a * (re_h_z_alpha_integrand + re_h_z_alpha_integrand_old);
                im_z_result += 0.5 * d_a * (im_h_z_alpha_integrand + im_h_z_alpha_integrand_old);
            }
        }
        re_integral_x[0] = re_x_result;
        im_integral_x[0] = im_x_result;
        re_integral_y[0] = re_y_result;
        im_integral_y[0] = im_y_result;
        re_integral_z[0] = re_z_result;
        im_integral_z[0] = im_z_result;
    }
    
    
    __global__ void magnetic_field(const float *re_kp_array, const float *im_kp_array, 
                                   const float *re_kz_array, const float *im_kz_array,
                                   const float *alpha_array,  
                                   const float *x_array, const float *y_array, const float *z_array,
                                   const float *re_g_te_array, const float *im_g_te_array,
                                   const float *re_g_tm_array, const float *im_g_tm_array,
                                   float *re_h_x, float *im_h_x, float *re_h_y, float *im_h_y, 
                                   float *re_h_z, float *im_h_z)
    {
        unsigned int i = blockIdx.x * blockDim.x + threadIdx.x;
        if (i >= LEN_X) return;
        
        float x = x_array[i];
        float y = y_array[i];
        float z = z_array[i];
        
        float re_kp = 0.0;
        float im_kp = 0.0;

        float re_x_result = 0.0;
        float im_x_result = 0.0;
        float re_y_result = 0.0;
        float im_y_result = 0.0;
        float re_z_result = 0.0;
        float im_z_result = 0.0;
        
        float re_x_kappa_integrand = 0.0;
        float im_x_kappa_integrand = 0.0;
        float re_y_kappa_integrand = 0.0;
        float im_y_kappa_integrand = 0.0;
        float re_z_kappa_integrand = 0.0;
        float im_z_kappa_integrand = 0.0;
        
        for (int i_k=0; i_k<LEN_K; i_k++)
        {
            float re_kp_old = re_kp;
            float im_kp_old = im_kp;
            
            re_kp = re_kp_array[i_k];
            im_kp = im_kp_array[i_k];
            
            float re_d_kp = re_kp - re_kp_old;
            float im_d_kp = im_kp - im_kp_old;

            float re_kz = re_kz_array[i_k];
            float im_kz = im_kz_array[i_k];
            
            float re_x_kappa_integrand_old = re_x_kappa_integrand;
            float im_x_kappa_integrand_old = im_x_kappa_integrand;
            float re_y_kappa_integrand_old = re_y_kappa_integrand;
            float im_y_kappa_integrand_old = im_y_kappa_integrand;
            float re_z_kappa_integrand_old = re_z_kappa_integrand;
            float im_z_kappa_integrand_old = im_z_kappa_integrand;
            
            float re_x_alpha_itgrl = 0.0;
            float im_x_alpha_itgrl = 0.0;
            float re_y_alpha_itgrl = 0.0;
            float im_y_alpha_itgrl = 0.0;
            float re_z_alpha_itgrl = 0.0;
            float im_z_alpha_itgrl = 0.0;
            
            alpha_integral(i_k, x, y, z, re_kp, im_kp, re_kz, im_kz, alpha_array, 
                           re_g_te_array, im_g_te_array, re_g_tm_array, im_g_tm_array,
                           &re_x_alpha_itgrl, &im_x_alpha_itgrl, &re_y_alpha_itgrl, &im_y_alpha_itgrl, 
                           &re_z_alpha_itgrl, &im_z_alpha_itgrl);
                           
            re_x_kappa_integrand = re_x_alpha_itgrl * re_kp - im_x_alpha_itgrl * im_kp;
            im_x_kappa_integrand = re_x_alpha_itgrl * im_kp + im_x_alpha_itgrl * re_kp;
            re_y_kappa_integrand = re_y_alpha_itgrl * re_kp - im_y_alpha_itgrl * im_kp;
            im_y_kappa_integrand = re_y_alpha_itgrl * im_kp + im_y_alpha_itgrl * re_kp;
            re_z_kappa_integrand = re_z_alpha_itgrl * re_kp - im_z_alpha_itgrl * im_kp;
            im_z_kappa_integrand = re_z_alpha_itgrl * im_kp + im_z_alpha_itgrl * re_kp;
            
            if (i_k>0)
            {
                re_x_result += 0.5 * (re_d_kp * (re_x_kappa_integrand + re_x_kappa_integrand_old) 
                                      - im_d_kp * (im_x_kappa_integrand + im_x_kappa_integrand_old));
                im_x_result += 0.5 * (re_d_kp * (im_x_kappa_integrand + im_x_kappa_integrand_old) 
                                      + im_d_kp * (re_x_kappa_integrand + re_x_kappa_integrand_old));
                re_y_result += 0.5 * (re_d_kp * (re_y_kappa_integrand + re_y_kappa_integrand_old) 
                                      - im_d_kp * (im_y_kappa_integrand + im_y_kappa_integrand_old));
                im_y_result += 0.5 * (re_d_kp * (im_y_kappa_integrand + im_y_kappa_integrand_old) 
                                      + im_d_kp * (re_y_kappa_integrand + re_y_kappa_integrand_old));
                re_z_result += 0.5 * (re_d_kp * (re_z_kappa_integrand + re_z_kappa_integrand_old) 
                                      - im_d_kp * (im_z_kappa_integrand + im_z_kappa_integrand_old));
                im_z_result += 0.5 * (re_d_kp * (im_z_kappa_integrand + im_z_kappa_integrand_old) 
                                      + im_d_kp * (re_z_kappa_integrand + re_z_kappa_integrand_old));
            }
        }
        re_h_x[i] = re_x_result;
        im_h_x[i] = im_x_result;
        re_h_y[i] = re_y_result;
        im_h_y[i] = im_y_result;
        re_h_z[i] = re_z_result;
        im_h_z[i] = im_z_result;
    }
"""


"""This subpackage contains functionality that has to do with the
representation of electromagnetic fields in spherical or plane vector wave
functions.
The __init__ module contains some helper functions (e.g. with respect to
SVWF indexing) and is the place to store default coordinate arrays for
Sommerfeld integrals and field expansions."""

import numpy as np
from numba import jit


def angular_frequency(vacuum_wavelength):
    """Angular frequency :math:`\omega = 2\pi c / \lambda`

    Args:
        vacuum_wavelength (float): Vacuum wavelength in length unit

    Returns:
        Angular frequency in the units of c=1 (time units=length units).
        This is at the same time the vacuum wavenumber.
    """
    return 2 * np.pi / vacuum_wavelength


###############################################################################
#                         SWE indexing                                        #
###############################################################################

@jit(nopython=True, cache=True)
def multi_to_single_index(tau, l, m, l_max, m_max):
    r"""Unique single index for the totality of indices characterizing a svwf expansion coefficient.

    The mapping follows the scheme:

    +-------------+-------------+-------------+------------+
    |single index |    spherical wave expansion indices    |
    +=============+=============+=============+============+
    | :math:`n`   | :math:`\tau`| :math:`l`   | :math:`m`  |
    +-------------+-------------+-------------+------------+
    |     1       |     1       |      1      |   -1       |
    +-------------+-------------+-------------+------------+
    |     2       |     1       |      1      |    0       |
    +-------------+-------------+-------------+------------+
    |     3       |     1       |      1      |    1       |
    +-------------+-------------+-------------+------------+
    |     4       |     1       |      2      |   -2       |
    +-------------+-------------+-------------+------------+
    |     5       |     1       |      2      |   -1       |
    +-------------+-------------+-------------+------------+
    |     6       |     1       |      2      |    0       |
    +-------------+-------------+-------------+------------+
    |    ...      |    ...      |     ...     |   ...      |
    +-------------+-------------+-------------+------------+
    |    ...      |     1       |    l_max    |    m_max   |
    +-------------+-------------+-------------+------------+
    |    ...      |     2       |      1      |   -1       |
    +-------------+-------------+-------------+------------+
    |    ...      |    ...      |     ...     |   ...      |
    +-------------+-------------+-------------+------------+

    Args:
        tau (int):      Polarization index :math:`\tau` (0=spherical TE, 1=spherical TM)
        l (int):        Degree :math:`l` (1, ..., lmax)
        m (int):        Order :math:`m` (-min(l,mmax),...,min(l,mmax))
        l_max (int):    Maximal multipole degree
        m_max (int):    Maximal multipole order

    Returns:
        single index (int) subsuming :math:`(\tau, l, m)`
    """
    # use:
    # \sum_{l=1}^lmax (2\min(l,mmax)+1) = \sum_{l=1}^mmax (2l+1) + \sum_{l=mmax+1}^lmax (2mmax+1)
    #                                   = 2*(1/2*mmax*(mmax+1))+mmax  +  (lmax-mmax)*(2*mmax+1)
    #                                   = mmax*(mmax+2)               +  (lmax-mmax)*(2*mmax+1)
    tau_blocksize = m_max * (m_max + 2) + (l_max - m_max) * (2 * m_max + 1)
    n = tau * tau_blocksize
    if (l - 1) <= m_max:
        n += (l - 1) * (l - 1 + 2)
    else:
        n += m_max * (m_max + 2) + (l - 1 - m_max) * (2 * m_max + 1)
    n += m + min(l, m_max)
    return n


@jit(nopython=True, cache=True)
def blocksize(l_max, m_max):
    """Number of coefficients in outgoing or regular spherical wave expansion for a single particle.

    Args:
        l_max (int):    Maximal multipole degree
        m_max (int):    Maximal multipole order
    Returns:
         Number of indices for one particle, which is the maximal index plus 1."""
    return multi_to_single_index(tau=1, l=l_max, m=m_max, l_max=l_max, m_max=m_max) + 1


###############################################################################
#                        PWE wave vectors                                     #
###############################################################################

def k_z(k_parallel=None, n_effective=None, k=None, omega=None, vacuum_wavelength=None, refractive_index=None):
    """z-component :math:`k_z=\sqrt{k^2-\kappa^2}` of the wavevector. The branch cut is defined such that the imaginary
    part is not negative, compare section 2.3.1 of [Egel 2018 dissertation].
    Not all of the arguments need to be specified.

    Args:
        k_parallel (numpy ndarray):     In-plane wavenumber :math:`\kappa` (inverse length)
        n_effective (numpy ndarray):    Effective refractive index :math:`n_\mathrm{eff}`
        k (float):                      Wavenumber (inverse length)
        omega (float):                  Angular frequency :math:`\omega` or vacuum wavenumber (inverse length, c=1)
        vacuum_wavelength (float):      Vacuum wavelength :math:`\lambda` (length)
        refractive_index (complex):     Refractive index :math:`n_i` of material

    Returns:
        z-component :math:`k_z` of wavenumber with non-negative imaginary part (inverse length)
    """
    if k_parallel is None:
        if omega is None:
            omega = angular_frequency(vacuum_wavelength)
        k_parallel = n_effective * omega

    if k is None:
        if omega is None:
            omega = angular_frequency(vacuum_wavelength)
        k = refractive_index * omega

    kz = np.sqrt(k ** 2 - k_parallel ** 2 + 0j)
    kz = (kz.imag >= 0) * kz + (kz.imag < 0) * (-kz)  # Branch cut such to prohibit negative imaginary
    return kz


"""The default arrays for k_parrallel, azimuthal_angles and polar_angles are
used in Sommerfeld integrals or in plane wave expansions whenever no other
arrays for the specification of the wavevectors are explicitly stated."""

def angular_arrays(angular_resolution=np.pi/360):
    """ Return azimuthal and polar angular arrays with a certain angular resolution """
    default_azimuthal_angles = np.arange(0, 2*np.pi+0.5*angular_resolution, angular_resolution, dtype=float)
    default_polar_angles = np.arange(0, np.pi+0.5*angular_resolution, angular_resolution, dtype=float)
    return default_azimuthal_angles, default_polar_angles

""" Default arrays for angular coordinates """
default_azimuthal_angles = None
default_polar_angles = None

"""Default n_effective array for Sommerfeld integrals - needs to be set, e.g. at beginning of simulation"""
default_Sommerfeld_k_parallel_array = None

"""Default n_effective array for the initial field (beams, dipoles) - needs to be set, e.g. at beginning of simulation"""
default_initial_field_k_parallel_array = None


def reasonable_neff_waypoints(layer_refractive_indices=None, neff_imag=1e-2, neff_max=None, neff_max_offset=1):
    """Construct a reasonable list of waypoints for a k_parallel array of plane wave expansions.
    The waypoints mark a contour through the complex plane such that possible waveguide mode and
    branchpoint singularity locations are avoided (see section 3.10.2.1 of [Egel 2018 dissertation]).

    Args:
        layer_refractive_indices (list or array):       Complex refractive indices of the plane layer system
        neff_imag (float):                              Extent of the contour into the negative imaginary direction
                                                        (in terms of effective refractive index, n_eff=kappa/omega).
        neff_max (float):                               Truncation value of contour (in terms of effective refractive
                                                        index).
        neff_max_offset (float):                        If no value for `neff_max` is specified, use the last estimated
                                                        singularity location plus this value (in terms of effective
                                                        refractive index). Default=1

    Returns:
        List of complex waypoint values.
    """
    if layer_refractive_indices is None:
        layer_refractive_indices = [1, 2]

    # estimate range of possible waveguide mode positions
    min_waveguide_neff = max(0, min(np.array(layer_refractive_indices).real) - 0.1)
    max_waveguide_neff = max(np.array(layer_refractive_indices).real) + 0.2

    if neff_max is None:
        neff_max = max_waveguide_neff + neff_max_offset

    if neff_imag == 0:
        return [0, neff_max]

    else:
        # avoid the real axis on a region between n_min-0.1 and n_max+0.2
        # this should save us from branchpoints and numerically critical waveguide modes
        # (SPPs can be outside of that interval but are then strongly damped)
        waypoints = [0,
                     min_waveguide_neff,
                     min_waveguide_neff - 1j * neff_imag,
                     max_waveguide_neff - 1j * neff_imag,
                     max_waveguide_neff]
        if neff_max > max_waveguide_neff:
            waypoints.append(neff_max)
        return waypoints


def create_neff_array(neff_waypoints, neff_resolution):
    """Construct an array of complex effective refractive index values. The effective refractive index is a
    dimensionless quantity that will be multiplied by vacuum wavenumber to yield the in-plane component of a wave vector.
    This is used for the plane wave expansion of fields and for Sommerfeld integrals. Complex contours are used to
    improve numerical stability (see section 3.10.2.1 of [Egel 2018 dissertation]).

    Args:
        neff_waypoints (list or ndarray): Corner points through which the contour runs
        neff_resolution(float):           Resolution of contour (i.e., distance between adjacent elements)

    Returns:
        Array of complex effective refractive index values
    """
    neff_segments = []
    for i in range(len(neff_waypoints) - 1):
        abs_dneff = abs(neff_waypoints[i + 1] - neff_waypoints[i])
        if abs_dneff > 0:
            num_samples = int(np.ceil(abs_dneff / neff_resolution)) + 1
            array = np.linspace(0, 1, num=num_samples, endpoint=True, dtype=complex)
            neff_segments.append(neff_waypoints[i] + array  * (neff_waypoints[i + 1] - neff_waypoints[i]))
    return np.concatenate(neff_segments)


def create_k_parallel_array(vacuum_wavelength, neff_waypoints, neff_resolution):
    """Construct an array of complex in-plane wavenumbers (i.e., the radial component of the cylindrical coordinates of
    the wave-vector). This is used for the plane wave expansion of fields and for Sommerfeld integrals.
    Complex contours are used to improve numerical stability
    (see section 3.10.2.1 of [Egel 2018 dissertation]).

    Args:
        vacuum_wavelength (float):        Vacuum wavelength :math:`\lambda` (length)
        neff_waypoints (list or ndarray): Corner points through which the contour runs
                                          This quantity is dimensionless (effective
                                          refractive index, will be multiplied by vacuum
                                          wavenumber)
        neff_resolution(float):           Resolution of contour, again in terms of
                                          effective refractive index

    Returns:
        Array :math:`\kappa_i` of in-plane wavenumbers (inverse length)
    """
    return create_neff_array(neff_waypoints, neff_resolution) * angular_frequency(vacuum_wavelength)


def branchpoint_correction(layer_refractive_indices, n_effective_array, neff_minimal_branchpoint_distance):
    """Check if an array of complex effective refractive index values (e.g. for Sommerfeld integration) contains
    possible branchpoint singularities and if so, replace them by nearby non-singular locations.

    Args:
        layer_refractive_indices (list or array):       Complex refractive indices of planarly layered medium
        n_effective_array (1d numpy.array):             Complex effective refractive indexc values that are to be checked
                                                        for branchpoint collision
                                                        This array is changed during the function evaluation!
        neff_minimal_branchpoint_distance (float):      Minimal distance that contour points shall have from
                                                        branchpoint singularities

    Returns:
        corrected n_effective_array
    """
    for n in layer_refractive_indices:
        while True:
            branchpoint_indices = np.where(abs(n_effective_array - n) < neff_minimal_branchpoint_distance)[0]
            if len(branchpoint_indices) == 0:
                break
            idx = branchpoint_indices[0]
            # replace contour point by two points at the middle towards its left and right neighbors
            if not idx == len(n_effective_array) - 1:
                n_effective_array = np.insert(n_effective_array,
                                              idx + 1,
                                              (n_effective_array[idx] + n_effective_array[idx+1]) / 2.0)
                # make sure the new point is ok, otherwise remove
                if abs(n_effective_array[idx + 1] - n) < neff_minimal_branchpoint_distance:
                    n_effective_array = np.delete(n_effective_array, idx + 1)
            if not idx == 0:
                n_effective_array[idx] = (n_effective_array[idx-1] + n_effective_array[idx]) / 2.0
                # make sure the shifted point is ok, otherwise remove
                if abs(n_effective_array[idx] - n) < neff_minimal_branchpoint_distance:
                    n_effective_array = np.delete(n_effective_array, idx)
    return n_effective_array


def reasonable_Sommerfeld_neff_contour(neff_waypoints=None, layer_refractive_indices=None, neff_imag=1e-2,
                                       neff_max=None, neff_max_offset=1, neff_resolution=1e-2,
                                       neff_minimal_branchpoint_distance=None):
    """
    Return a reasonable n_effective array that is suitable for the construction of a Sommerfeld k_parallel integral
    contour. Use this function if you don't want to care for numerical details of your simulation.

    Args:
        neff_waypoints (list or ndarray): Corner points through which the contour runs
                                          This quantity is dimensionless (effective
                                          refractive index, will be multiplied by vacuum
                                          wavenumber)
                                          If not provided, reasonable waypoints are estimated.
        layer_refractive_indices (list):  Complex refractive indices of planarly layered medium
                                          Only needed when no neff_waypoints are provided
        neff_imag (float):                Extent of the contour into the negative imaginary direction
                                          (in terms of effective refractive index, n_eff=kappa/omega).
                                          Only needed when no neff_waypoints are provided
        neff_max (float):                 Truncation value of contour (in terms of effective refractive index).
                                          Only needed when no neff_waypoints are provided
        neff_max_offset (float):          Use the last estimated singularity location plus this value (in terms of
                                          effective refractive index). Default=1
                                          Only needed when no neff_waypoints are provided and if no value for `neff_max`
                                          is specified.
        neff_resolution(float):           Resolution of contour, again in terms of effective refractive index
        neff_minimal_branchpoint_distance (float):      Minimal distance that contour points shall have from
                                                        branchpoint singularities (in terms of effective refractive
                                                        index). This is only relevant if not deflected into imaginary.
                                                        Default: One fifth of neff_resolution

    Returns:
        Array of complex effective refractive index values

    """
    if neff_waypoints is None:
        neff_waypoints = reasonable_neff_waypoints(layer_refractive_indices=layer_refractive_indices,
                                                   neff_imag=neff_imag,
                                                   neff_max=neff_max,
                                                   neff_max_offset=neff_max_offset)
    if neff_minimal_branchpoint_distance is None:
        neff_minimal_branchpoint_distance = neff_resolution / 5
    n_effective = create_neff_array(neff_waypoints, neff_resolution)
    if layer_refractive_indices is not None:
        n_effective = branchpoint_correction(layer_refractive_indices=layer_refractive_indices,
                                             n_effective_array=n_effective,
                                             neff_minimal_branchpoint_distance=neff_minimal_branchpoint_distance)
    return n_effective


def reasonable_Sommerfeld_kpar_contour(vacuum_wavelength, neff_waypoints=None, layer_refractive_indices=None,
                                        neff_imag=1e-2, neff_max=None, neff_max_offset=1, neff_resolution=1e-2,
                                        neff_minimal_branchpoint_distance=None):
    """
    Return a reasonable k_parallel array that is suitable as a Sommerfeld integral contour.
    Use this function if you don't want to care for numerical details of your simulation.

    Args:
        vacuum_wavelength (float):        Vacuum wavelength :math:`\lambda` (length)
        neff_waypoints (list or ndarray): Corner points through which the contour runs
                                          This quantity is dimensionless (effective
                                          refractive index, will be multiplied by vacuum
                                          wavenumber)
                                          If not provided, reasonable waypoints are estimated.
        layer_refractive_indices (list):  Complex refractive indices of planarly layered medium
                                          Only needed when no neff_waypoints are provided
        neff_imag (float):                Extent of the contour into the negative imaginary direction
                                          (in terms of effective refractive index, n_eff=kappa/omega).
                                          Only needed when no neff_waypoints are provided
        neff_max (float):                 Truncation value of contour (in terms of effective refractive index).
                                          Only needed when no neff_waypoints are provided
        neff_max_offset (float):          Use the last estimated singularity location plus this value (in terms of
                                          effective refractive index). Default=1
                                          Only needed when no neff_waypoints are provided and if no value for `neff_max`
                                          is specified.
        neff_resolution(float):           Resolution of contour, again in terms of effective refractive index
        neff_minimal_branchpoint_distance (float):      Minimal distance that contour points shall have from
                                                        branchpoint singularities (in terms of effective refractive
                                                        index). This is only relevant if not deflected into imaginary.
                                                        Default: One fifth of neff_resolution

    Returns:
        Array :math:`\kappa_i` of in-plane wavenumbers (inverse length)

    """
    n_effective = reasonable_Sommerfeld_neff_contour(neff_waypoints=neff_waypoints,
                                                     layer_refractive_indices=layer_refractive_indices,
                                                     neff_imag=neff_imag,
                                                     neff_max=neff_max,
                                                     neff_max_offset=neff_max_offset,
                                                     neff_resolution=neff_resolution,
                                                     neff_minimal_branchpoint_distance=neff_minimal_branchpoint_distance)
    return n_effective * angular_frequency(vacuum_wavelength)



class ProcessesCluster:
    def __init__(self, max_processes):
        self.processes = []
        self.max_processes = max_processes

    def can_add_new_process(self):
        return len(self.processes) < self.max_processes

    def add_process(self, process):
        self.processes.append(process)

    def execute(self):
        for p in self.processes:
            p.start()

        for p in self.processes:
            p.join()


def distribute_processes_into_clusters(processes, max_processes_in_cluster):
    clusters = []
    cluster = ProcessesCluster(max_processes_in_cluster)
    clusters.append(cluster)

    for p in processes:
        if not cluster.can_add_new_process():
            cluster = ProcessesCluster(max_processes_in_cluster)
            clusters.append(cluster)

        cluster.add_process(p)

    return clusters


"""This module contains functions to evaluate the coupling between a 
single particle and a periodic particle arrangement."""

import smuthi.periodicboundaries.ewald_lattice_sums as pbels
import smuthi.periodicboundaries.ewald_helper as pbeh
import smuthi.periodicboundaries.coupling_helper as pbch
import smuthi.fields as flds
import smuthi.utility.math as sf
import smuthi.linearsystem.particlecoupling.layer_mediated_coupling as laycoup
from numba import njit, prange
import numpy as np

@njit()
def discrete_scattering_angles(k, k0t, azimuthal_angle_init, a1, a2):
    """Evaluates the discrete scattering angles of a plane wave exciting a periodic
    particle arrangement. 
    In particular, combinations of in-plane wavenumber and azimuthal angle. 
    Args:
        k (float):                      initial field's wavenumber
        k0t (numpy.ndarray):            complex in-plane wave vector in Carthesian coordinates 
        azimuthal_angle_init (float):   azimuthal angle of the initial field
        a1 (numpy.ndarray):             lattice vector 1 in Carthesian coordinates
        a2 (numpy.ndarray):             lattice vector 2 in Carthesian coordinates
    Returns:
        k_parallel (numpy.ndarray):         in-plane wavenumbers
        azimuthal_angles (numpy.ndarray):   azimuthal angles
        kpar_alpha_comb (list):             list of length len(k_parallel) that connects indices 
                                            of k_parallel with indices of azimuthal_angles 
        gmax (int):                         maximal order of unit cells to which the reciprocal
                                            lattive vector g is evaluated (k = k_inc - g) 
    """
    b1, b2 = pbeh.reciprocal_lattice_vec(a1, a2) 
    
     # find all kpar / alpha combinations and form a matrix
    kpar_list = np.array([np.linalg.norm(k0t)], np.float64)
    alpha_list = np.array([azimuthal_angle_init], np.float64)
    kpar_alpha_list = [np.array([0], np.int64)]
      
    flag, num_evan = True, 0
    gmax = 0
    # ensures that all propagating scattering orders + 3 evanescent orders are taken into account
    # 3 evanescent orders are somewhat arbitrary (could be set from outside)
    while flag == True or num_evan < 3:
        gmax += 1
        n1, n2 = pbeh.n1n2_indices(gmax)
        g = n1.reshape(n1.shape[0], 1) * b1.reshape(1, 2) + n2.reshape(n2.shape[0], 1) * b2.reshape(1, 2)
        ksca = k0t.reshape(2) - g    
        
        kpar = np.zeros(ksca.shape[0], np.float64)
        alpha = np.zeros(ksca.shape[0], np.float64)
        for idx in range(ksca.shape[0]):
            kpar[idx] = np.linalg.norm(ksca[idx, :])
            alpha[idx] = np.arctan2(ksca[idx, 1], ksca[idx, 0]) 
        for idx, ang in enumerate(alpha):
            if ang < 0:
                alpha[idx] = 2 * np.pi + ang
            
        if np.min(kpar) > k.real: # all scattered waves are evanescent
            flag = False
            num_evan += 1
        
        for idx in range(kpar.shape[0]):     
            isclose_a = np.zeros(alpha_list.shape[0], np.int32)
            for ii in range(alpha_list.shape[0]):
                isclose_a[ii] = np.round(alpha[idx], 5) == np.round(alpha_list[ii], 5)
            if not isclose_a.any():
                alpha_list = np.append(alpha_list, alpha[idx])
                alpha_idx = alpha_list.shape[0] - 1
            else:
                alpha_idx = np.where(isclose_a == 1)[0][0] 
                
            isclose_k = np.zeros(kpar_list.shape[0], np.int32)
            for ii in range(kpar_list.shape[0]):
                isclose_k[ii] = np.round(kpar[idx], 5) == np.round(kpar_list[ii], 5)
            if not isclose_k.any():
                kpar_list = np.append(kpar_list, kpar[idx])
                kpar_alpha_list.append(np.array([alpha_idx], np.int64))  
            else:
                kpar_idx = np.where(isclose_k == 1)[0][0]
                kpar_alpha_list[kpar_idx] = np.append(kpar_alpha_list[kpar_idx], alpha_idx)
       
    k_parallel = np.sort(kpar_list)
    azimuthal_angles = np.sort(alpha_list)
    
    for idx, kpar in enumerate(k_parallel):
        kpar_idx = np.where(kpar == kpar_list)[0][0]
        for idx_2, alpha_idx in enumerate(kpar_alpha_list[kpar_idx]):
            alpha = alpha_list[alpha_idx]
            alpha_idx_new = np.where(alpha == azimuthal_angles)[0][0]           
            if idx_2 == 0:
                arr = [alpha_idx_new]
            else:
                arr.append(alpha_idx_new)
        arr.sort()        
        if idx == 0:
            kpar_alpha_comb = [arr]
        else:
            kpar_alpha_comb.append(arr)

    return k_parallel, azimuthal_angles, kpar_alpha_comb, gmax


@njit(parallel=True)
def periodic_coupling_matrix(vacuum_wavelength, k0t, azimuthal_angle, layer_thicknesses, layer_refractive_indices,
                            i_sca, positions, radii, lmax_array, mmax_array, a1, a2, eta, a5b5_mat, mmax_global):
    """ Coupling matrix between one particle and a periodic particle arrangement.
        "prange()" enables the multithreading of the Numba code. 
    Args:
        vacuum_wavelength (float):          vacuum wavelength in length unit   
        k0t (numpy.ndarray):                complex in-plane wave vector in Carthesian coordinates  
        azimuthal_angle (numpy.ndarray):    azimuthal angle of excitation 
        layer_thicknesses (numpy.ndarray):  layer thicknesses of stratified medium
        layer_refractive_indices (numpy.ndarray): layer refractive indices of stratified medium
        i_sca (int):                        layer number that contains all scattering particles
        positions (numpy.ndarray):          position of each particle
        radii (numpy.ndarray):              radius of each particle
        lmax_array (numpy.ndarray):         lmax of each particle 
        mmax_array (numpy.ndarray):         mmax of each particle
        a1 (numpy.ndarray):                 lattice vector 1 in Carthesian coordinates
        a2 (numpy.ndarray):                 lattice vector 2 in Carthesian coordinates
        eta (float):                        Ewald sum separation parameter
        a5b5_mat (numpy.ndarray):           a5, b5 coefficient lookup table
        mmax_global (int):                  global maximal multipole order of all particles 
    Returns:
        sum of direct and layer mediated coupling matrices
    """    
    s_max = positions.shape[0]
    shape = pbch.index_block(s_max - 1, lmax_array, mmax_array)[1]    
    coup_mat = np.zeros((shape, shape), np.complex128)
    
    # alternative to status bar, to access the progress of the prange-numba-loop
    progress = np.zeros((s_max, s_max), np.int64)
    sections = np.zeros(10, np.float64) # in 10% steps
    print('Coupling matrix computation finished: 0.0 %.')
    
    for s1 in prange(s_max):
        idx1 = pbch.index_block(s1, lmax_array, mmax_array)
        for s2 in prange(s_max):
            idx2 = pbch.index_block(s2, lmax_array, mmax_array)
            coup_mat[idx1[0]:idx1[1], idx2[0]:idx2[1]] = (
                                    periodic_direct_coupling_block(vacuum_wavelength, layer_refractive_indices[i_sca],
                                            k0t, azimuthal_angle, positions[s1], lmax_array[s1], mmax_array[s1], radii[s1],
                                            positions[s2], lmax_array[s2], mmax_array[s2], radii[s2], a1, a2, eta, a5b5_mat, mmax_global)
                                    + periodic_layer_mediated_coupling_block(vacuum_wavelength, layer_thicknesses,
                                              layer_refractive_indices, i_sca, k0t, azimuthal_angle, positions[s1], lmax_array[s1],
                                              mmax_array[s1], radii[s1], positions[s2], lmax_array[s2], mmax_array[s2],
                                              radii[s2], a1, a2))
            progress[s1, s2] = 1 
            if (np.sum(progress) / s_max ** 2) >= np.sum(sections) + 0.1:
                sections[:int(np.floor(np.sum(progress) / s_max ** 2 * 1e1))] = 0.1
                print('Coupling matrix computation finished:', np.round(np.sum(sections) * 100), '%.')
                
    return coup_mat



@njit()
def periodic_direct_coupling_block(vacuum_wavelength, refractive_index, k0t, azimuthal_angle, 
                                   rs1, l_max1, m_max1, radius1, rs2, l_max2, m_max2, radius2,
                                   a1, a2, eta, a5b5_lookup, mmax_global):
    """  Direct particle coupling matrix :math:`W^R` between one (receiving) particle and
        a (emitting) periodic particle arrangement.
        Based on a intermediate plane wave expansion or an Ewald lattice sum. 
    Args:
        vacuum_wavelength (float):          vacuum wavelength in length unit   
        refractive_index (complex):         refractive index of the scattering layer 
        k0t (numpy.ndarray):                complex in-plane wave vector in Carthesian coordinates  
        azimuthal_angle (numpy.ndarray):    azimuthal angle of excitation 
        rs1 (numpy.ndarray):                position of the receiving particle in Carthesian coordinates 
        l_max1 (int):                       maximal multipole degree of the receiving particle
        m_max1 (int):                       maximal multipole order of the receiving particle
        radius1 (float):                    radius of the circumscribing sphere of the receiving particle
        rs2 (numpy.ndarray):                position of the emitting particle in Carthesian coordinates 
        l_max2 (int):                       maximal multipole degree of the emitting particle
        m_max2 (int):                       maximal multipole order of the emitting particle
        radius2 (float):                    radius of the circumscribing sphere of the emitting particle
        a1 (numpy.ndarray):                 lattice vector 1 in Carthesian coordinates
        a2 (numpy.ndarray):                 lattice vector 2 in Carthesian coordinates
        eta (float):                        Ewald sum separation parameter
        a5b5_lookup (numpy.ndarray):        lookup table for a5, b5 coefficients
        mmax_global (int):                  global maximal multipole order of all particles 
    
    Returns:
        w (numpy.ndarray):                  direct coupling block 
    """
    blocksize1 = flds.blocksize(l_max1, m_max1)
    blocksize2 = flds.blocksize(l_max2, m_max2)
    w = np.zeros((blocksize1, blocksize2), np.complex128)
    
    k = complex(2 * np.pi * refractive_index / vacuum_wavelength)
    
    # intermediate plane wave expansion
    if np.abs(rs1[2] - rs2[2]) > radius1 + radius2: # no vertical overlap of particles' circumscribing spheres
        k_parallel, azimuthal_angles, kpar_alpha_comb, gmax = discrete_scattering_angles(k, k0t, azimuthal_angle, a1, a2) 
        
        rs2s1 = rs1 - rs2
        rhos2s1 = np.linalg.norm(rs2s1[0:2])
        phis2s1 = np.arctan2(rs2s1[1], rs2s1[0])
        
        kz = np.sqrt(k ** 2 - k_parallel ** 2 + 0j)
        kz = (kz.imag >= 0) * kz + (kz.imag < 0) * (-kz)  # Branch cut to prohibit negative imaginary part
        
        # transformation coefficients; either upwards or downwards
        B = [np.zeros((2, blocksize1, k_parallel.shape[0]), np.complex128),
             np.zeros((2, blocksize2, k_parallel.shape[0]), np.complex128)]
        # list index: particle, np indices: pol, n, kpar_idx
        
        m_vec = [np.zeros(blocksize1, np.int32), np.zeros(blocksize2, np.int32)]
        
        # precompute spherical functions
        ct = kz / k
        st = k_parallel / k
        if rs2s1[2] > 0:
            _, pilm, taulm = sf.legendre_normalized_numbed(ct, st, l_max1)
        else:
            _, pilm, taulm = sf.legendre_normalized_numbed(-ct, st, l_max1)
            
        for tau in range(2):
            for m in range(-m_max1, m_max1 + 1):
                for l in range(max(1, abs(m)), l_max1 + 1):
                    n = flds.multi_to_single_index(tau, l, m, l_max1, m_max1)
                    m_vec[0][n] = m
                    for pol in range(2):
                        B[0][pol, n, :] = pbch.transformation_coefficients_vwf(tau, l, m, pol, kp=k_parallel, kz=kz,
                                                                          pilm=pilm, taulm=taulm, dagger=True)                  
        if rs2s1[2] > 0:
            _, pilm, taulm = sf.legendre_normalized_numbed(ct, st, l_max2)
        else:
            _, pilm, taulm = sf.legendre_normalized_numbed(-ct, st, l_max2)
                         
        for tau in range(2):
            for m in range(-m_max2, m_max2 + 1):
                for l in range(max(1, abs(m)), l_max2 + 1):
                    n = flds.multi_to_single_index(tau, l, m, l_max2, m_max2)
                    m_vec[1][n] = m
                    for pol in range(2):
                        B[1][pol, n, :] = pbch.transformation_coefficients_vwf(tau, l, m, pol, kp=k_parallel, kz=kz,
                                                                          pilm=pilm, taulm=taulm, dagger=False)
                        
        # k_parallel (and idx), azimuthal_angles arrays that reduce the summation to non-zero elements 
        alpha_vec = np.zeros((2 * gmax + 1) ** 2, np.float64)
        kpar_vec = np.zeros((2 * gmax + 1) ** 2, np.complex128)
        k_idx = np.zeros((2 * gmax + 1) ** 2, np.int32)
        s = 0
        for idx in range(k_parallel.shape[0]):
            num = len(kpar_alpha_comb[idx])
            alpha_vec[s:s+num] = azimuthal_angles[np.array(kpar_alpha_comb[idx], np.int32)]
            kpar_vec[s:s+num] = k_parallel[idx]
            k_idx[s:s+num] = idx
            s += num
            
        kz_vec = np.sqrt(k ** 2 - kpar_vec ** 2 + 0j)
        kz_vec = (kz_vec.imag >= 0) * kz_vec + (kz_vec.imag < 0) * (-kz_vec)  # Branch cut  to prohibit negative imaginary part
        jacobi_vector = 1 / (kz_vec * k)
        if rs2s1[2] < 0:
            kz_vec = -kz_vec
            
        prefac = 8 * np.pi / np.linalg.norm(np.cross(a1, a2))
        m2_minus_m1 = m_vec[1] - m_vec[0].reshape(m_vec[0].shape[0], 1)
        eikrs2s1= np.exp(1j * (kpar_vec * rhos2s1 * np.cos(alpha_vec - phis2s1) + kz_vec * rs2s1[2]))
        
        for n1 in range(blocksize1):
            for n2 in range(blocksize2):
                BB = np.zeros(k_parallel.shape[0], np.complex128) 
                eimpma = np.exp(1j * m2_minus_m1[n1, n2] * alpha_vec)
                for pol in range(2):
                    BB += B[0][pol, n1, :] * B[1][pol, n2, :]
                w[n1, n2] = np.sum(eimpma * eikrs2s1 * jacobi_vector * BB[k_idx])               
        w = prefac * w 
    
    # Ewald lattice sum 
    else:
        c = rs1 - rs2 
        for l1 in range(1, l_max1 + 1):
            for l2 in range(1, l_max2 + 1):
                for m1 in range(-min(m_max1, l1), min(m_max1, l1) + 1):
                    for m2 in range(-min(m_max2, l2), min(m_max2, l2) + 1):
                        M = m2 - m1
                        prefac = pbeh.normalization_spherical_harmonics(M)
                        A, B = complex(0), complex(0) 
                        for L in range(max(abs(l1 - l2), abs(M)), l1 + l2 + 1): # if L < abs(M) then P=0 
                            Dlm = pbels.D_LM(L, M, k, k0t, a1, a2, eta, c)
                            A += a5b5_lookup[0, l1 - 1, l2 - 1, m1 + mmax_global, m2 + mmax_global, L] * Dlm
                            B += a5b5_lookup[1, l1 - 1, l2 - 1, m1 + mmax_global, m2 + mmax_global, L] * Dlm
                        for tau1 in range(2):
                            n1 = flds.multi_to_single_index(tau1, l1, m1, l_max1, m_max1)
                            for tau2 in range(2):
                                n2 = flds.multi_to_single_index(tau2, l2, m2, l_max2, m_max2)
                                if tau1 == tau2:
                                    w[n1, n2] = prefac * A
                                else:
                                    w[n1, n2] = prefac * B 
    return w


@njit()
def periodic_layer_mediated_coupling_block(vacuum_wavelength, layer_thicknesses, layer_refractive_indices,
                                           i_sca, k0t, azimuthal_angle, rs1, l_max1, m_max1, radius1,
                                           rs2, l_max2, m_max2, radius2, a1, a2):
    """ Layer-system mediated particle coupling matrix :math:`W^R` between one (receiving) particle
        and a (emitting) periodic particle arrangement.
    Args:
        vacuum_wavelength (float):          vacuum wavelength in length unit
        layer_thicknesses (numpy.ndarray):  layer thicknesses of stratified medium
        layer_refractive_indices (numpy.ndarray): layer refractive indices of stratified medium
        i_sca (int):                        layer number that contains all scattering particles
        k0t (numpy.ndarray):                complex in-plane wave vector in Carthesian coordinates  
        azimuthal_angle (numpy.ndarray):    azimuthal angle of excitation 
        rs1 (numpy.ndarray):                position of the receiving particle in Carthesian coordinates 
        l_max1 (int):                       maximal multipole degree of the receiving particle
        m_max1 (int):                       maximal multipole order of the receiving particle
        radius1 (float):                    radius of the circumscribing sphere of the receiving particle
        rs2 (numpy.ndarray):                position of the emitting particle in Carthesian coordinates 
        l_max2 (int):                       maximal multipole degree of the emitting particle
        m_max2 (int):                       maximal multipole order of the emitting particle
        radius2 (float):                    radius of the circumscribing sphere of the emitting particle
        a1 (numpy.ndarray):                 lattice vector 1 in Carthesian coordinates
        a2 (numpy.ndarray):                 lattice vector 2 in Carthesian coordinates
        a5b5_mat (numpy.ndarray):           a5, b5 coefficient lookup table
        mmax_global (int):                  global maximal multipole order of all particles 
    
    Returns:
        wr (numpy.ndarray):                 layer mediated coupling block 
    """
    blocksize1 = flds.blocksize(l_max1, m_max1)
    blocksize2 = flds.blocksize(l_max2, m_max2)
    wr = np.zeros((blocksize1, blocksize2), np.complex128)
    
    omega = 2 * np.pi / vacuum_wavelength
    k = complex(omega * layer_refractive_indices[i_sca])
    k_parallel, azimuthal_angles, kpar_alpha_comb, gmax = discrete_scattering_angles(k, k0t, azimuthal_angle, a1, a2)
    kz = np.sqrt(k ** 2 - k_parallel ** 2 + 0j)
    kz = (kz.imag >= 0) * kz + (kz.imag < 0) * (-kz)  # Branch cut to prohibit negative imaginary part
    
    rs2s1 = rs1 - rs2
    rhos2s1 = np.linalg.norm(rs2s1[0:2])
    phis2s1 = np.arctan2(rs2s1[1], rs2s1[0])
    ziss1 = rs1[2] - np.sum(layer_thicknesses[:i_sca])
    ziss2 = rs2[2] - np.sum(layer_thicknesses[:i_sca])
    
    # phase factors
    ejkz = np.zeros((2, 2, k_parallel.shape[0]), np.complex128)  # indices are: particle, plus/minus, kpar_idx
    ejkz[0, 0, :] = np.exp(1j * kz * ziss1)
    ejkz[0, 1, :] = np.exp(- 1j * kz * ziss1)
    ejkz[1, 0, :] = np.exp(1j * kz * ziss2)
    ejkz[1, 1, :] = np.exp(- 1j * kz * ziss2)
    
    # remove inf from ejkz
    # (otherwise evanescent waves cause overflow in case of very thick layers)
    idces =  np.where(np.isinf(ejkz[0, 1, :]))[0]
    for ii in idces:
        ejkz[0, 1, ii] = 1.79769313e+308
    idces =  np.where(np.isinf(ejkz[1, 1, :]))[0]
    for ii in idces:
        ejkz[1, 1, ii] = 1.79769313e+308
    
    # layer response
    L = np.zeros((2, 2, 2, k_parallel.shape[0]), np.complex128)  # polarization, pl/mn1, pl/mn2, kpar_idx
    for pol in range(2):
        L[pol, :, :, :] = pbch.layersystem_response_matrix(pol, layer_thicknesses, layer_refractive_indices,
                                                      k_parallel, omega, i_sca, i_sca)
        
    # transformation coefficients; either upwards or downwards
    B = [np.zeros((2, 2, blocksize1, k_parallel.shape[0]), np.complex128),
         np.zeros((2, 2, blocksize2, k_parallel.shape[0]), np.complex128)]
    # list index: particle, np indices: pol, plus/minus, n, kpar_idx
    
    m_vec = [np.zeros(blocksize1, np.int32), np.zeros(blocksize2, np.int32)]
    
    # precompute spherical functions
    ct = kz / k
    st = k_parallel / k    
    _, pilm_list_pl, taulm_list_pl = sf.legendre_normalized_numbed(ct, st, l_max1)
    _, pilm_list_mn, taulm_list_mn = sf.legendre_normalized_numbed(-ct, st, l_max1)
    pilm = (pilm_list_pl, pilm_list_mn)
    taulm = (taulm_list_pl, taulm_list_mn)
    
    for tau in range(2):
        for m in range(-m_max1, m_max1 + 1):
            for l in range(max(1, abs(m)), l_max1 + 1):
                n = flds.multi_to_single_index(tau, l, m, l_max1, m_max1)
                m_vec[0][n] = m
                for iplmn in range(2):
                    for pol in range(2):
                        B[0][pol, iplmn, n, :] = pbch.transformation_coefficients_vwf(tau, l, m, pol, kp=k_parallel, kz=kz,
                                                                                 pilm=pilm[iplmn],
                                                                                 taulm=taulm[iplmn], dagger=True)

    _, pilm_list_pl, taulm_list_pl = sf.legendre_normalized_numbed(ct, st, l_max2)
    _, pilm_list_mn, taulm_list_mn = sf.legendre_normalized_numbed(-ct, st, l_max2)
    pilm = (pilm_list_pl, pilm_list_mn)
    taulm = (taulm_list_pl, taulm_list_mn)
                       
    for tau in range(2):
        for m in range(-m_max2, m_max2 + 1):
            for l in range(max(1, abs(m)), l_max2 + 1):
                n = flds.multi_to_single_index(tau, l, m, l_max2, m_max2)
                m_vec[1][n] = m
                for iplmn in range(2):
                    for pol in range(2):
                        B[1][pol, iplmn, n, :] = pbch.transformation_coefficients_vwf(tau, l, m, pol, kp=k_parallel, kz=kz,
                                                                                 pilm=pilm[iplmn],
                                                                                 taulm=taulm[iplmn], dagger=False)

    # k_parallel (and idx), azimuthal_angles arrays that reduce the summation to non-zero elements 
    alpha_vec = np.zeros((2 * gmax + 1) ** 2, np.float64)
    kpar_vec = np.zeros((2 * gmax + 1) ** 2, np.complex128)
    k_idx = np.zeros((2 * gmax + 1) ** 2, np.int32)
    s = 0
    for idx in range(k_parallel.shape[0]):
        num = len(kpar_alpha_comb[idx])
        alpha_vec[s:s+num] = azimuthal_angles[np.array(kpar_alpha_comb[idx], np.int32)]
        kpar_vec[s:s+num] = k_parallel[idx]
        k_idx[s:s+num] = idx
        s += num
        
    prefac = 8 * np.pi / np.linalg.norm(np.cross(a1, a2))
    jacobi_vector = 1 / (kz * k)
    m2_minus_m1 = m_vec[1] - m_vec[0].reshape(m_vec[0].shape[0], 1)
    eikparrs2s1 = np.exp(1j * kpar_vec * rhos2s1 * np.cos(alpha_vec - phis2s1))
    
    for n1 in range(blocksize1):
        BeL = np.zeros((2, 2, k_parallel.shape[0]), np.complex128)  # indices are: pol, plmn2, n1, kpar_idx
        for iplmn1 in range(2):
            for pol in range(2):
                BeL[pol, :, :] += (L[pol, iplmn1, :, :]
                                        * B[0][pol, iplmn1, n1, :]
                                        * ejkz[0, iplmn1, :])
        for n2 in range(blocksize2):
            eimpma = np.exp(1j * m2_minus_m1[n1, n2] * alpha_vec)
            BeLBe = np.zeros((k_parallel.shape[0]), np.complex128)
            laycoup.eval_BeLBe(BeLBe, BeL, B[1], ejkz, n2)           
            wr[n1, n2] = np.sum(eimpma * eikparrs2s1 * jacobi_vector[k_idx] * BeLBe[k_idx])  
    return prefac * wr    